/*
 * deviced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/types.h>
#include <vconf.h>

#include "hbm.h"
#include "lbm.h"
#include "auto-brightness-sensorhub.h"
#include "display/util.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "core/device-notifier.h"
#include "battery/power-supply.h"
#include "shared/plugin.h"

#define LOWBATCAPACITY 5

static struct display_plugin *disp_plgn;
static struct _backlight_ops *backlight_ops;
static int auto_brightness_state = SETTING_BRIGHTNESS_AUTOMATIC_OFF;
static bool lbm, hbm, hold_brt, lowdim;

static void change_brightness_transit(int start, int end)
{
	struct display_config *display_conf = get_var_display_config();
	if(!display_conf) {
		_E("Failed to get display configuration variable.");
		return ;
	}
	backlight_ops->transit_brt(start, end,
			display_conf->brightness_change_step);
}

static void set_brightness_level(int level)
{
	int default_brightness;
	int current_brightness;

	if (get_pm_cur_state() != S_NORMAL)
		return;
/*
	if (battery.capacity <= LOWBATCAPACITY && battery.charge_now != CHARGER_CHARGING)
		return;
*/
	_I("level changed! %d", level);

	if ((level > PM_MAX_BRIGHTNESS || level < PM_MIN_BRIGHTNESS) &&
	    (level != LBM_LEVEL && level != HBM_LEVEL && level != SPECIFIC_MODE_OFF)) {
		_E("Invalid level %d", level);
		return;
	}

	switch (level) {
	case LBM_LEVEL:
		/*
		 * received LBM enable
		 * change brightness to under LBM_BRIGHTNESS_LOW and set lbm state
		 */
		auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		auto_brightness_control(BR_LBM_ON, BR_IMPLICIT);
		break;
	case HBM_LEVEL:
		/* HBM must do not turn on at DIM_STAY state. */
		if (get_pm_status_flag() & DIM_MASK)
			break;
		/*
		 * received HBM enable
		 * change brightness to MAX value and set hbm state
		 */
		auto_brightness_control(BR_LBM_OFF, BR_IMPLICIT);
		auto_brightness_control(BR_HBM_ON, BR_IMPLICIT);
		break;
	case SPECIFIC_MODE_OFF:
		default_brightness = backlight_ops->get_default_brt();
		backlight_ops->get_brightness(&current_brightness);

		/* Disable HBM, LBM */
		auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		auto_brightness_control(BR_LBM_OFF, BR_IMPLICIT);

		/* For AOD -> LCDON, brightness effect is deferred to this point */
		if (current_brightness != default_brightness && !hold_brt && !lowdim) {
			change_brightness_transit(current_brightness, default_brightness);
			auto_brightness_control(BR_SET_BRIGHTNESS, default_brightness);
		}

		break;
	default:
		/*
		 * received automatic brightness by the light sensor
		 */
		if (auto_brightness_state != SETTING_BRIGHTNESS_AUTOMATIC_ON)
			return;

		auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		auto_brightness_control(BR_LBM_OFF, BR_IMPLICIT);
		auto_brightness_control(BR_SET_BRIGHTNESS, level);
		break;
	}
}

static void set_default_brightness(void)
{
	int default_brt;

	default_brt = backlight_ops->get_default_brt();
	backlight_ops->set_default_brt(default_brt);
	backlight_ops->update();
}

static void set_automatic_state_cb(keynode_t *key_nodes, void *data)
{
	if (key_nodes == NULL) {
		_E("Wrong parameter, key_nodes is null.");
		return;
	}
	auto_brightness_state = vconf_keynode_get_int(key_nodes);

	switch (auto_brightness_state) {
	case SETTING_BRIGHTNESS_AUTOMATIC_OFF:
		auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		/* escape dim state if it's in low battery.*/
		set_brightness_changed_state();
		set_default_brightness();
		break;
	case SETTING_BRIGHTNESS_AUTOMATIC_PAUSE:
		auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		break;
	case SETTING_BRIGHTNESS_AUTOMATIC_ON:
		set_brightness_changed_state();
		break;
	default:
		_E("Invalid value %d.", auto_brightness_state);
	}
}

static void change_flag(bool *flag, bool value)
{
	// previous
	int plbm = lbm;
	int phbm = hbm;
	int phold = hold_brt;
	int pld = lowdim;

	if (flag != &lbm && flag != &hbm && flag != &hold_brt && flag != &lowdim) {
		_E("Invalid flag setting.");
		return;
	}

	if (*flag == value)
		return;

	// update
	*flag = value;

	_D("Change flag LBM/HBM/Hold/LowDim: %d%d%d%d -> %d%d%d%d",
		plbm, phbm, phold, pld,
		lbm, hbm, hold_brt, lowdim);
}

int auto_brightness_control(enum brightness_request_e request, int set_brightness)
{
	int default_brightness;
	int current_brightness;
	int ret;

	default_brightness = backlight_ops->get_default_brt();
	backlight_ops->get_brightness(&current_brightness);

	if (request == BR_LBM_ON) {
		if (!get_lbm_setting())
			return 0;
		if (!lbm && !hbm && !hold_brt && !lowdim) {
			change_brightness_transit(default_brightness, lbm_down_brt(default_brightness));
			lbm_set_state(true);
			backlight_ops->set_brightness(default_brightness);
			change_flag(&lbm, 1);
		} else if (!lbm && !hbm && !hold_brt && lowdim) {
			change_brightness_transit(current_brightness, lbm_down_brt(current_brightness));
			lbm_set_state(true);
			backlight_ops->set_brightness(current_brightness);
			change_flag(&lbm, 1);
		} else if (!lbm && !hbm && hold_brt) {
			change_flag(&lbm, 1);
		}
	} else if (request == BR_HBM_ON) {
		if (!lbm && !hbm && !hold_brt) {
			if (!lowdim)
				change_brightness_transit(default_brightness, PM_MAX_BRIGHTNESS);
			else
				change_brightness_transit(current_brightness, PM_MAX_BRIGHTNESS);
			hbm_set_state(true);
			change_flag(&hbm, 1);
		} else if (!lbm && !hbm && hold_brt) {
			_E("Someone holds brightness, HBM is ignored.");
			return -EBUSY;
		}
	} else if (request == BR_HOLD_BRIGHTNESS) {
		if (hbm) {
			_E("HBM state. Hold brightness is ignored.");
			return -EBUSY;
		}
		if (lbm && !hold_brt) {
			lbm_set_state(false);
			change_brightness_transit(lbm_down_brt(current_brightness), set_brightness);
		} else {
			backlight_ops->set_brightness(set_brightness);
		}
		change_flag(&hold_brt, 1);
	} else if (request == BR_LOWDIM_ON) {
		if (!lowdim) {
			if (!hbm && !hold_brt && (get_pm_cur_state() == S_NORMAL))
				backlight_ops->dim();
			change_flag(&lowdim, 1);
		}
	} else if (request == BR_LBM_OFF) {
		if (lbm && !hold_brt) {
			lbm_set_state(false);
			change_flag(&lbm, 0);

			/* called by DEVICE_NOTIFIER_LCD_OFF_COMPLETE, no need to change brightness */
			if (get_pm_cur_state() != S_NORMAL)
				return 0;

			if (!lowdim) {
				change_brightness_transit(lbm_down_brt(default_brightness), default_brightness);
				backlight_ops->set_brightness(default_brightness);
			} else {
				backlight_ops->dim();
			}
		} else if (lbm && hold_brt) {
			change_flag(&lbm, 0);
		}
	} else if (request == BR_HBM_OFF) {
		if (hbm) {
			hbm_set_state(false);
			if (!lowdim)
				change_brightness_transit(PM_MAX_BRIGHTNESS, default_brightness);
			else
				change_brightness_transit(PM_MAX_BRIGHTNESS, PM_DIM_BRIGHTNESS);
			change_flag(&hbm, 0);
		}
	} else if (request == BR_RELEASE_BRIGHTNESS) {
		if (!lbm && hold_brt) {
			change_flag(&hold_brt, 0);

			/* called by DEVICE_NOTIFIER_LCD_OFF_COMPLETE, no need to change brightness */
			if (get_pm_cur_state() != S_NORMAL)
				return 0;

			if (!lowdim)
				backlight_ops->set_brightness(default_brightness);
			else
				backlight_ops->dim();
		} else if (lbm && hold_brt) {
			change_flag(&hold_brt, 0);
			if (!lowdim) {
				change_brightness_transit(current_brightness, lbm_down_brt(default_brightness));
				lbm_set_state(true);
				backlight_ops->set_brightness(default_brightness);
			} else {
				change_brightness_transit(current_brightness, lbm_down_brt(PM_DIM_BRIGHTNESS));
				lbm_set_state(true);
				backlight_ops->set_brightness(PM_DIM_BRIGHTNESS);
			}
		}
	} else if (request == BR_LOWDIM_OFF) {
		if (lowdim) {
			if (!hbm && !hold_brt && (get_pm_cur_state() == S_NORMAL))
				backlight_ops->update();
			change_flag(&lowdim, 0);
		}
	} else if (request == BR_SET_BRIGHTNESS) {
		if (lbm && hold_brt)
			lbm_set_state(true);
		change_flag(&lowdim, 0);
		change_flag(&hold_brt, 0);
		backlight_ops->set_default_brt(set_brightness);
		if (!hbm) {
			ret = backlight_ops->set_brightness(set_brightness);
			if (ret < 0) {
				_E("Failed to set brightness to %d.", set_brightness);
				return ret;
			}
		}
		ret = vconf_set_int(VCONFKEY_SETAPPL_LCD_BRIGHTNESS, set_brightness);
		if (ret < 0)
			_E("Failed to set vconf value for lcd brightness: %d", vconf_get_ext_errno());
	} else {
		_E("Invalid request, %d", request);
		return -EINVAL;
	}

	return 0;
}

static int display_off_changed(void *data)
{
	auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);

	return 0;
}

static int display_off_complete_changed(void *data)
{
	auto_brightness_control(BR_LBM_OFF, BR_IMPLICIT);
	auto_brightness_control(BR_RELEASE_BRIGHTNESS, BR_IMPLICIT);

	return 0;
}

/* restore brightness on recovering battery health to normal */
int auto_brightness_restore(void)
{
	/* maintain hold brightness */
	if (hold_brt)
		return 0;

	/* update to the default brightness */
	return backlight_ops->update();
}

int prepare_level_handler(void)
{
	int status, ret;

	display_info.set_brightness_level = set_brightness_level;
	disp_plgn->auto_brightness_control = auto_brightness_control;

	ret = vconf_get_int(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT, &status);
	if (ret >= 0)
		auto_brightness_state = status;

	vconf_notify_key_changed(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT,
				 set_automatic_state_cb, NULL);

	register_notifier(DEVICE_NOTIFIER_LCD_OFF, display_off_changed);
	register_notifier(DEVICE_NOTIFIER_LCD_OFF_COMPLETE, display_off_complete_changed);

	return 0;
}

void exit_level_handler(void)
{
	vconf_ignore_key_changed(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT,
				 set_automatic_state_cb);
	set_default_brightness();

	unregister_notifier(DEVICE_NOTIFIER_LCD_OFF, display_off_changed);
	unregister_notifier(DEVICE_NOTIFIER_LCD_OFF_COMPLETE, display_off_complete_changed);
}

static void __CONSTRUCTOR__ initialize(void)
{
	disp_plgn = get_var_display_plugin();
	if (!disp_plgn)
		_E("Failed to get display plugin variable.");

	backlight_ops = get_var_backlight_ops();
	if (!backlight_ops)
		_E("Failed to get backlight operator variable.");
}
