/*
 * deviced
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <hw/bezel.h>
#include <assert.h>
#include <vconf.h>

#include "core/devices.h"
#include "core/common.h"
#include "core/device-notifier.h"
#include "display/util.h"
#include "display/poll.h"
#include "display/core.h"
#include "display-ops.h"
#include "shared/plugin.h"

#ifndef VCONFKEY_SETAPPL_WAKEUP_BY_BEZEL_ENABLE
#define VCONFKEY_SETAPPL_WAKEUP_BY_BEZEL_ENABLE "db/setting/wakeup_by_bezel_enable"
#endif

#ifndef VCONFKEY_SETAPPL_TOUCH_BEZEL_FEEDBACK_VIBRATION_LEVEL_INT
#define VCONFKEY_SETAPPL_TOUCH_BEZEL_FEEDBACK_VIBRATION_LEVEL_INT "db/setting/touch_bezel/feedback/vibration_level"
#endif

enum bezel_type {
	BEZEL_NOTSUP = 0,
	BEZEL_HARD,
	BEZEL_SOFT,
};

static enum bezel_state bezel_wakeup_control(void);

static struct display_plugin *disp_plgn;
static struct bezel_device *bezel_dev;
static int bezel_wakeup = 0;
static int theater_mode = 0;
static int goodnight_mode = 0;
static enum bezel_type bezel_type = BEZEL_NOTSUP;

static void bezel_changed_cb(keynode_t *key_nodes, void *data)
{
	assert(key_nodes);

	bezel_wakeup = vconf_keynode_get_bool(key_nodes);

	_I("Bezel wakeup condition: %d", bezel_wakeup);
	if (disp_plgn->pm_change_internal)
		disp_plgn->pm_change_internal(INTERNAL_LOCK_PM, LCD_NORMAL);

	device_notify(DEVICE_NOTIFIER_BEZEL_WAKEUP, (void *)((intptr_t)bezel_wakeup));
}

static void theater_changed_cb(keynode_t *key_nodes, void *data)
{
	enum bezel_state state;

	assert(key_nodes);

	if (!bezel_dev || !bezel_dev->set_state) {
		_E("bezel_dev->set_state is not supported.");
		return;
	}

	theater_mode = vconf_keynode_get_bool(key_nodes);
	_I("Theater mode: %d.", theater_mode);

	if (get_pm_cur_state() == LCD_OFF) {
		state = bezel_wakeup_control();
		bezel_dev->set_state(state);
	}
}

static void goodnight_changed_cb(keynode_t *key_nodes, void *data)
{
	enum bezel_state state;

	assert(key_nodes);

	if (!bezel_dev || !bezel_dev->set_state) {
		_E("bezel_dev->set_state is not supported.");
		return;
	}

	goodnight_mode = vconf_keynode_get_bool(key_nodes);
	_I("Goodnight mode: %d.", goodnight_mode);

	if (get_pm_cur_state() == LCD_OFF) {
		state = bezel_wakeup_control();
		bezel_dev->set_state(state);
	}
}

static void bezel_rotary_event_cb(keynode_t *key_nodes, void *data)
{
	enum bezel_state bezel_state;
	int ret;

	assert(key_nodes);

	bezel_state = (enum bezel_state) vconf_keynode_get_bool(key_nodes);

	if (!bezel_dev || !bezel_dev->set_sw_state) {
		_E("bezel_dev->set_sw_state is not supported.");
		return;
	}

	ret = bezel_dev->set_sw_state(bezel_state);

	if (ret == 0) {
		_I("Bezel rotary event condition is %d", bezel_state);
		if (disp_plgn->pm_change_internal)
			disp_plgn->pm_change_internal(INTERNAL_LOCK_PM, LCD_NORMAL);
	} else {
		_E("Failed to set bezel rotary event condition.");
	}
}

static void bezel_rotary_vibration_set_state(int state)
{
	int ret;

	if (!bezel_dev || !bezel_dev->set_vib_state) {
		_E("bezel_dev->set_vib_state is not supported.");
		return;
	}

	ret = bezel_dev->set_vib_state(state);
	if (ret < 0)
		_E("Can't set vib state %d (ret %d)", state, ret);
}

static void bezel_rotary_vibration_changed_cb(keynode_t *key_nodes, void *data)
{
	int bezel_vib_state;

	assert(key_nodes);

	bezel_vib_state = vconf_keynode_get_int(key_nodes);

	_I("bezel rotary vibration state is %d", bezel_vib_state);

	bezel_rotary_vibration_set_state(bezel_vib_state);
}

static int bezel_probe(void *data)
{
	struct hw_info *info;
	int ret;

	if (bezel_dev)
		return 0;

	ret = hw_get_info(BEZEL_HARDWARE_DEVICE_ID, (const struct hw_info **)&info);
	if (ret < 0) {
		_I("bezel shared library is not supported: %d", ret);
		return -ENOTSUP;
	}

	if (!info->open) {
		_E("Fail to open bezel device : open(NULL)");
		return -ENODEV;
	}

	ret = info->open(info, NULL, (struct hw_common **)&bezel_dev);
	if (ret < 0) {
		_E("Fail to get bezel device structure: %d", ret);
		return ret;
	}

	return 0;
}

static void bezel_init(void *data)
{
	int ret;
	int bezel_rotary_event = -1;
	int bezel_vib_state = -1;
	enum bezel_state init_bezel_state;

	if (!bezel_dev)
		return;

	if (bezel_dev->get_sw_state && bezel_dev->get_sw_state(&init_bezel_state) == 0) {
		_I("Init SW bezel, state: %d", init_bezel_state);
		bezel_type = BEZEL_SOFT;
	} else if (bezel_dev->get_state && bezel_dev->get_state(&init_bezel_state) == 0) {
		_I("Init HW bezel, state: %d", init_bezel_state);
		bezel_type = BEZEL_HARD;
	} else {
		_E("Failed to init bezel.");
		return;
	}


	if (bezel_type == BEZEL_HARD) {
		vconf_notify_key_changed(VCONFKEY_SETAPPL_WAKEUP_BY_BEZEL_ENABLE, bezel_changed_cb, NULL);
		vconf_notify_key_changed(VCONFKEY_SETAPPL_THEATER_MODE_ENABLE, theater_changed_cb, NULL);
		vconf_notify_key_changed(VCONFKEY_SETAPPL_GOODNIGHT_MODE_ENABLE, goodnight_changed_cb, NULL);

		ret = vconf_get_bool(VCONFKEY_SETAPPL_WAKEUP_BY_BEZEL_ENABLE, &bezel_wakeup);
		if (ret == 0)
			_I("Bezel wakeup condition: %d.", bezel_wakeup);
		else
			_E("Failed to get vconf value for wakeup by bezel enable: %d.", vconf_get_ext_errno());

		ret = vconf_get_bool(VCONFKEY_SETAPPL_THEATER_MODE_ENABLE, &theater_mode);
		if (ret == 0)
			_I("Theater mode: %d.", theater_mode);
		else
			_E("Failed to get vconf value for theater mode: %d.", vconf_get_ext_errno());

		ret = vconf_get_bool(VCONFKEY_SETAPPL_GOODNIGHT_MODE_ENABLE, &goodnight_mode);
		if (ret == 0)
			_I("Goodnight mode: %d.", goodnight_mode);
		else
			_E("Failed to get vconf value for goodnight mode: %d.", vconf_get_ext_errno());
	} else if (bezel_type == BEZEL_SOFT) {
		vconf_notify_key_changed(VCONFKEY_SETAPPL_ROTARY_EVENT_ENABLED_BOOL, bezel_rotary_event_cb, NULL);
		vconf_notify_key_changed(VCONFKEY_SETAPPL_TOUCH_BEZEL_FEEDBACK_VIBRATION_LEVEL_INT,
			bezel_rotary_vibration_changed_cb, NULL);

		ret = vconf_get_bool(VCONFKEY_SETAPPL_ROTARY_EVENT_ENABLED_BOOL, &bezel_rotary_event);
		if (ret == 0) {
			ret = bezel_dev->set_sw_state(bezel_rotary_event);
			if (ret == 0)
				_I("Bezel rotary event conditoin: %d.", bezel_rotary_event);
			else
				_E("Failed to set sw bezel state, %d.", ret);
		} else {
			_E("Failed to get vconf value for rotary event by bezen enabled: %d.", vconf_get_ext_errno());
		}

		ret = vconf_get_int(VCONFKEY_SETAPPL_TOUCH_BEZEL_FEEDBACK_VIBRATION_LEVEL_INT, &bezel_vib_state);
		if (ret == 0)
			bezel_rotary_vibration_set_state(bezel_vib_state);
		else
			_E("Failed to get vconf value for touch bezel feedback vibration level: %d", vconf_get_ext_errno());
	}
	_I("bezel condition (wakeup:%d) (rotary_event:%d) (bezel_vib_state %d)", bezel_wakeup, bezel_rotary_event, bezel_vib_state);
}

static void bezel_exit(void *data)
{
	struct hw_info *info;

	if (!bezel_dev)
		return;

	info = bezel_dev->common.info;

	assert(info);

	info->close((struct hw_common *)bezel_dev);
}

static enum bezel_state bezel_wakeup_control(void)
{
	if (!bezel_wakeup || theater_mode || goodnight_mode)
		return BEZEL_TURNOFF;
	else
		return BEZEL_TURNON;
}

static int bezel_start(enum device_flags flags)
{
	if (!bezel_dev)
		return 0;

	if (bezel_type == BEZEL_HARD)
		return bezel_dev->set_state(BEZEL_TURNON);

	return 0;
}

static int bezel_stop(enum device_flags flags)
{
	enum bezel_state state;

	if (!bezel_dev)
		return 0;

	if (bezel_type == BEZEL_HARD) {
		state = bezel_wakeup_control();
		return bezel_dev->set_state(state);
	}

	return 0;
}

/*
 * 'bezel' device_ops should be initialize before init key-filter.
 */
static const struct device_ops bezel_device_ops = {
	.priority = DEVICE_PRIORITY_HIGH,
	DECLARE_NAME_LEN("bezel"),
	.probe    = bezel_probe,
	.init     = bezel_init,
	.exit     = bezel_exit,
	.start    = bezel_start,
	.stop     = bezel_stop,
};

DEVICE_OPS_REGISTER(&bezel_device_ops)

static void __CONSTRUCTOR__ initialize(void)
{
	disp_plgn = get_var_display_plugin();
	if (!disp_plgn)
		_E("Failed to get display plugin variable.");
}
