/*
 * deviced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <math.h>

#include "auto-brightness-sensorhub.h"
#include "lbm.h"
#include "display-info.h"
#include "display/util.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "core/common.h"
#include "core/config-parser.h"
#include "core/device-notifier.h"

#define DISPLAY_CONF_FILE	"/etc/deviced/display.conf"

#define SIGNAL_LBM_ON	"LBMOn"
#define SIGNAL_LBM_OFF	"LBMOff"

#define LCD_PHASED_DELAY		10000   /* microsecond */

#define LBM_TRANSIT_STEP		20

static struct lbm_config lbm_conf = {
	.support = 0,
	.down_level = 0,
	.brt_table_size = 0,
	.brt_table = NULL,
	.aod_brightness_level = 0,
};

static struct _backlight_ops *backlight_ops;
static int lbm_setting_mode;
static int system_brightness;
static struct display_device *display_dev;

static void broadcast_lbm_state(int state)
{
	char *str;

	if (state == 1)
		str = SIGNAL_LBM_ON;
	else
		str = SIGNAL_LBM_OFF;

	dbus_handle_emit_dbus_signal(NULL,
			DEVICED_PATH_DISPLAY,
			DEVICED_INTERFACE_DISPLAY,
			str,
			NULL);
}

static int get_level_by_brt(int brt)
{
	int iter;
	for (iter = 0; iter < lbm_conf.brt_table_size; iter++) {
		if (lbm_conf.brt_table[iter] == brt)
			return iter;
	}
	return -EINVAL;
}

int lbm_down_brt(int brt)
{
	int level;
	int down_level;

	if (lbm_conf.down_level == 0) {
		_I("LBM down level setting is 0.");
		return brt;
	}

	if (lbm_conf.brt_table == NULL || lbm_setting_mode == 0) {
		_I("LBM setting mode is %d.", lbm_setting_mode);
		return brt;
	}

	level = get_level_by_brt(brt);
	down_level = level - lbm_conf.down_level;

	if (down_level < 0)
		down_level = 0;

	return lbm_conf.brt_table[down_level];
}

static int lbm_get_brightness(int *val)
{
	*val = system_brightness;

	return 0;
}

static int lbm_set_brightness(int val)
{
	int brt = 0;

	if (!display_dev || !display_dev->set_brightness) {
		_E("There is no display device.");
		return -ENOENT;
	}

	if (get_pm_status_flag() & DIM_MASK)
		val = 0;
	else
		brt = lbm_down_brt(val);

	system_brightness = val;

	_I("Set brightness(LBM on) system=%d, real=%d.", val, brt);
	device_notify(DEVICE_NOTIFIER_DISPLAY_BRIGHTNESS, (void *)&val);

	return display_dev->set_brightness(brt);
}

static void lbm_change_brightness(int start, int end, int step)
{
	int diff, val;
	int ret = -1;
	int prev;

	if (display_dimstay_check())
		return;

	ret = backlight_ops->get_brightness(&prev);

	if (ret < 0) {
		_E("Failed to get brightness, %d.", ret);
		return;
	}

	if (prev == end)
		return;

	if (get_pm_status_flag() & DIM_MASK)
		end = 0;

	_I("Start %d, end %d, step %d.", start, end, step);

	if (display_dev && display_dev->set_multi_brightness) {
		diff = lbm_down_brt(end);

		ret = display_dev->set_multi_brightness(diff, step, LCD_PHASED_DELAY);
		if (ret < 0)
			_E("Failed to set_multi_brightness, %d.", ret);

		backlight_ops->set_brightness(end);

		return;
	}

	diff = end - start;

	if (abs(diff) < step)
		val = (diff > 0 ? 1 : -1);
	else
		val = (int)ceil((double)diff / step);

	while (start != end) {
		if (val == 0)
			break;

		start += val;
		if ((val > 0 && start > end) ||
		    (val < 0 && start < end))
			start = end;

		usleep(LCD_PHASED_DELAY);
		backlight_ops->set_brightness(start);
	}
}

int lbm_set_state(int lbm)
{
	_I("lbm_set_state, %d", lbm);

	broadcast_lbm_state(lbm);
	if (lbm) {
		backlight_ops->set_brightness = lbm_set_brightness;
		backlight_ops->get_brightness = lbm_get_brightness;
		backlight_ops->transit_brt = lbm_change_brightness;
	} else {
		backlight_ops->restore_brightness_func();
	}

	return 0;
}

int lbm_get_state(void)
{
	if (!lbm_conf.support)
		return -ENODEV;

	return backlight_ops->set_brightness == lbm_set_brightness;
}

static void lbm_table_load(char *value, struct lbm_config *c)
{
	char *p, *saveptr;
	int level_count = 1;
	int i;

	if (!value || value[0] == '\0')
		return;

	for (i = 0; *(value + i) != '\0'; i++) {
		if (*(value + i) == ',')
			level_count++;
	}

	c->brt_table = malloc(sizeof(int) * level_count);
	if (!c->brt_table) {
		_E("Failed to allocate memory.");
		return;
	}
	c->brt_table_size = level_count;

	i = 0;
	p = strtok_r(value, ",", &saveptr);
	if (p)
		c->brt_table[i++] = atoi(p);

	while (p != NULL) {
		p = strtok_r(NULL, ",", &saveptr);
		if (p)
			c->brt_table[i++] = atoi(p);
	}
}

static int lbm_load_config(struct parse_result *result, void *user_data)
{
	struct lbm_config *c = user_data;

	_D("%s,%s,%s.", result->section, result->name, result->value);

	if (!c)
		return -EINVAL;

	if (!MATCH(result->section, "Display"))
		return 0;

	if (MATCH(result->name, "LBMsupport")) {
		c->support = (MATCH(result->value, "yes") ? 1 : 0);
		_D("lbm support is %d", c->support);
	} else if (MATCH(result->name, "Level")) {
		SET_CONF(c->down_level, atoi(result->value));
		_D("lbm down level is %d", c->down_level);
	} else if (MATCH(result->name, "BrtTable")) {
		lbm_table_load(result->value, c);
		_D("LBM table loaded.");
	}

	return 0;
}

static void lbm_mode_changed(keynode_t *key_nodes, void *data)
{
	int mode = vconf_keynode_get_bool(key_nodes);

	_I("LBM setting value is %s.", mode ? "enable" : "disable");

	lbm_setting_mode = mode;

	auto_brightness_control(BR_LBM_OFF, BR_IMPLICIT);
}

bool get_lbm_setting(void)
{
	return lbm_setting_mode;
}

static GVariant *dbus_getlbm(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int lbm;

	lbm = lbm_get_state();
	if (lbm < 0)
		_E("Failed to get low brightness mode %d.", lbm);
	else
		_D("Get low brightness mode %d.", lbm);

	return g_variant_new("(i)", lbm);
}

static const dbus_method_s dbus_methods[] = {
	{ "GetLBM",          NULL,   "i", dbus_getlbm },
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_DISPLAY,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};
static void lbm_init(void *data)
{
	int ret;

	/* load configutation */
	ret = config_parse(DISPLAY_CONF_FILE, lbm_load_config, &lbm_conf);
	if (ret < 0)
		_W("Failed to load %s, %d. Use default value.", DISPLAY_CONF_FILE, ret);

	ret = vconf_get_bool(VCONFKEY_SETAPPL_ACCESSIBILITY_AUTOMATIC_LOW_BRIGHTNESS, &lbm_setting_mode);
	if (ret < 0)
		_E("Failed to get VCONFKEY_SETAPPL_ACCESSIBILITY_AUTOMATIC_LOW_BRIGHTNESS, %d.", ret);

	vconf_notify_key_changed(VCONFKEY_SETAPPL_ACCESSIBILITY_AUTOMATIC_LOW_BRIGHTNESS, lbm_mode_changed, NULL);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_DISPLAY, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method, %d.", ret);

	display_dev = display_dev_get();

	_I("LBM setting value is %d.", lbm_setting_mode);
}

static const struct display_ops display_lbm_ops = {
	.name     = "lbm",
	.init     = lbm_init,
};

DISPLAY_OPS_REGISTER(&display_lbm_ops)

static void __CONSTRUCTOR__ initialize(void)
{
	backlight_ops = get_var_backlight_ops();
	if (!backlight_ops)
		_E("Failed to get backlight operator variable.");
}
