#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <vconf.h>

#include "test.h"

#define CMD_INTERVAL_SEC		1	/* second */
#define CMD_LONG_INTERVAL_SEC	3	/* second */

#define print_command(command) \
	do {\
		printf("%12s | ", command); \
		fflush(stdout);	\
		sleep(CMD_INTERVAL_SEC); \
	} while(0)

extern void battery_scenario(char *unit, int status);

static bool tc_success;
static int success, fail;

static void start_udev(void)
{
	dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
			DEVICED_PATH_SYSNOTI,
			DEVICED_INTERFACE_SYSNOTI,
			"udev",
			g_variant_new("(sis)", "udev", 1, "start"));
}

static void stop_udev(void)
{
	dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
			DEVICED_PATH_SYSNOTI,
			DEVICED_INTERFACE_SYSNOTI,
			"udev",
			g_variant_new("(sis)", "udev", 1, "stop"));
}

static int call_display_method_sync(const char *method, GVariant *param)
{
	GVariant *msg;
	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
			DEVICED_PATH_DISPLAY,
			DEVICED_INTERFACE_DISPLAY,
			method, param);
	if (!msg) {
		_E("Failed to %s.", method);
		return -1;
	}

	return 0;
}

static void check_result(int expect_default, int expect_current)
{
	int result_default;
	int result_current;
	bool match_default;
	bool match_current;
	GVariant *msg;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
			DEVICED_PATH_DISPLAY,
			DEVICED_INTERFACE_DISPLAY,
			"GetBrightnessInfo", NULL);
	if (!msg) {
		_E("Failed to GetBrightnessInfo.");
		return;
	}

	if (!g_variant_get_safe(msg, "(ii)", &result_default, &result_current)) {
		_D("Return type mismatching from GetBrightnessInfo.");
		return;
	}

	match_default = expect_default == result_default;
	match_current = expect_current == result_current;

	printf("%6d %6d %c | %6d %6d %c\n",
		expect_default, result_default, match_default ? ' ' : 'X',
		expect_current, result_current, match_current ? ' ' : 'X');

	tc_success &= (match_default && match_current);
}

static void hbm_on(int expect_default, int expect_current)
{
	print_command("HBM ON");

	call_display_method_sync("AutoBrightnessChanged", g_variant_new("(i)", 120));
	check_result(expect_default, expect_current);
	sleep(CMD_INTERVAL_SEC);
}

static void hbm_off(int expect_default, int expect_current)
{
	print_command("HBM OFF");

	call_display_method_sync("AutoBrightnessChanged", g_variant_new("(i)", 0));
	check_result(expect_default, expect_current);
	sleep(CMD_INTERVAL_SEC);
}

static void lbm_on(int expect_default, int expect_current)
{
	print_command("LBM ON");

	call_display_method_sync("AutoBrightnessChanged", g_variant_new("(i)", 110));
	check_result(expect_default, expect_current);
	sleep(CMD_INTERVAL_SEC);
}

static void lbm_off(int expect_default, int expect_current)
{
	print_command("LBM OFF");

	call_display_method_sync("AutoBrightnessChanged", g_variant_new("(i)", 0));
	check_result(expect_default, expect_current);
	sleep(CMD_INTERVAL_SEC);
}

static void hold(int hold, int expect_default, int expect_current)
{
	char command[32];

	snprintf(command, sizeof(command), "HOLD %d", hold);
	print_command(command);

	call_display_method_sync("HoldBrightness", g_variant_new("(i)", hold));
	check_result(expect_default, expect_current);
	sleep(CMD_INTERVAL_SEC);
}

static void release(int expect_default, int expect_current)
{
	print_command("RELEASE");

	call_display_method_sync("ReleaseBrightness", NULL);
	check_result(expect_default, expect_current);
	sleep(CMD_INTERVAL_SEC);
}

static void lowdim_on(int expect_default, int expect_current)
{
	print_command("LOWDIM ON");

	battery_scenario("bat3", 1);

	check_result(expect_default, expect_current);
	sleep(CMD_LONG_INTERVAL_SEC);
}

static void lowdim_off(int expect_default, int expect_current)
{
	print_command("LOWDIM OFF");

	battery_scenario("bat3", 0);

	check_result(expect_default, expect_current);
	sleep(CMD_LONG_INTERVAL_SEC);
}

static void set(int brightness, int expect_default, int expect_current)
{
	char command[32];

	snprintf(command, sizeof(command), "SET %d", brightness);
	print_command(command);

	call_display_method_sync("SetBrightness", g_variant_new("(ii)", 0, brightness));
	sleep(CMD_INTERVAL_SEC);
	check_result(expect_default, expect_current);
}

static void abnormal_health_on(int expect_default, int expect_current)
{
	print_command("ABNORMAL ON");

	battery_scenario("heat1", 1);

	check_result(expect_default, expect_current);
	sleep(CMD_LONG_INTERVAL_SEC);
}

static void abnormal_health_off(int expect_default, int expect_current)
{
	print_command("ABNORMAL OFF");

	battery_scenario("heat1", 0);

	check_result(expect_default, expect_current);
	sleep(CMD_LONG_INTERVAL_SEC);
}

static void initialize_test_env(int brightness, const char *fname)
{
	printf("[%s]\n", fname);

	tc_success = true;

	call_display_method_sync("ChangeState", g_variant_new("(s)", "lcdoff"));
	sleep(CMD_INTERVAL_SEC);
	call_display_method_sync("ChangeState", g_variant_new("(s)", "lcdon"));
	sleep(CMD_LONG_INTERVAL_SEC);
	printf(" Init brightness: %3d\n", brightness);
	call_display_method_sync("SetBrightness", g_variant_new("(ii)", 0, brightness));
	sleep(CMD_INTERVAL_SEC);
	call_display_method_sync("lockstate", g_variant_new("(sssi)", "lcdon", "staycurstate", "NULL", 0));
	sleep(CMD_INTERVAL_SEC);

	printf("             |    default      |    current\n");
	printf("     Command | expect result   | expect result\n");
	sleep(CMD_INTERVAL_SEC);
}
#define initialize(brightness)	initialize_test_env(brightness, __func__)

#define DEFINE_TC(tcname) static void tc_##tcname(void)
DEFINE_TC(hbm)
{
	initialize(50);

	hbm_on(50, 100);
	hbm_off(50, 50);
}

DEFINE_TC(lbm)
{
	initialize(100);

	/* BrtTable=1,20,40,50,60,70,80,90,95,100
	 * Level=4
	 * It lowers brightness 4 level down */
	lbm_on(100, 70);
	set(1, 1, 1);
	set(20, 20, 1);
	set(40, 40, 1);
	set(50, 50, 1);
	set(60, 60, 1);
	set(70, 70, 20);
	set(80, 80, 40);
	set(90, 90, 50);
	set(95, 95, 60);
	set(100, 100, 70);
	lbm_off(100, 100);
}

DEFINE_TC(hold)
{
	initialize(50);

	hold(100, 50, 100);
	release(50, 50);
}

DEFINE_TC(set)
{
	initialize(100);

	set(40, 40, 40);
}

DEFINE_TC(lowdim)
{
	initialize(100);

	lowdim_on(100, 0);
	lowdim_off(100, 100);
}

DEFINE_TC(abnormal_health)
{
	initialize(100);

	abnormal_health_on(100, 0);
	abnormal_health_off(100, 100);
}

DEFINE_TC(custom1)
{
	initialize(50);

	lowdim_on(50, 0);
	hold(100, 50, 100);
	lbm_on(50, 100);
	lbm_off(50, 100);
	release(50, 0);
	lowdim_off(50, 50);
}

#define TESTCASE(tc_function)	{.tcname = #tc_function, .tc = tc_##tc_function }
struct tc_list {
	const char *tcname;
	void (*tc)(void);
} tc[] = {
	/* default testcase */
	TESTCASE(hbm),
	TESTCASE(lbm),
	TESTCASE(hold),
	TESTCASE(set),
	TESTCASE(lowdim),
	TESTCASE(abnormal_health),

	/* custom testcase */
	TESTCASE(custom1),
};

#define array_len(arr)	(sizeof(arr) / sizeof((arr)[0]))
static void do_test(const char *tcname)
{
	int i;
	int timeout;
	int ret = -1;

	ret = vconf_get_int(VCONFKEY_SETAPPL_LCD_TIMEOUT_NORMAL, &timeout);
	if (ret == 0) {
		if (vconf_set_int(VCONFKEY_SETAPPL_LCD_TIMEOUT_NORMAL, 300) < 0)
			printf("Failed to set LCDNORMAL timeout, %d.\n", vconf_get_ext_errno());
	}

	stop_udev();

	if (tcname) {
		for (i = 0; i < array_len(tc); ++i) {
			if (strcmp(tcname, tc[i].tcname))
				continue;
			tc[i].tc();
			if (tc_success)
				++success;
			else
				++fail;
			printf("--------------------------------------------------\n");
		}
	} else {
		for (i = 0; i < array_len(tc); ++i) {
			tc[i].tc();
			if (tc_success)
				++success;
			else
				++fail;
			printf("--------------------------------------------------\n");
		}
	}
	printf("Success: %d, Fail: %d\n", success, fail);

	start_udev();
	if (ret == 0) {
		if (vconf_set_int(VCONFKEY_SETAPPL_LCD_TIMEOUT_NORMAL, timeout) < 0)
			printf("Failed to restore LCDNORMAL timeout, %d.\n", vconf_get_ext_errno());
	}
}

static void brightness_init(void *data)
{
	do_test(NULL);
}

static int brightness_unit(int argc, char **argv)
{
	if (argc < 4)
		do_test(NULL);
	else
		do_test(argv[3]);

	return 0;
}

static void brightness_exit(void *data)
{
	_I("End brightness test.");
}

static const struct test_ops brightness_test_ops = {
	.priority = TEST_PRIORITY_NORMAL,
	.name     = "brightness",
	.init     = brightness_init,
	.exit    = brightness_exit,
	.unit    = brightness_unit,
};

TEST_OPS_REGISTER(&brightness_test_ops)
