/*
 * test
 *
 * Copyright (c) 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "test.h"

#define METHOD_IR_ISAVAILABLE		"IRIsAvailable"
#define METHOD_IR_TRANSMIT			"TransmitIR"
#define METHOD_IR_SETIRCOMMAND		"SetIrCommand"

struct dbus_int {
	int *list;
	int size;
};

static bool request_ir_method(const char *method, const char *sig, int *param)
{
	GVariant *msg;
	int val;
	bool ret = FALSE;
	GVariant *var = NULL;

	if (sig && g_variant_type_is_array(G_VARIANT_TYPE(sig))) {
		var = dbus_handle_make_simple_array(sig, param);
	} else if (!sig && !param) {
		var = NULL;
	} else {
		_E("wrong signature %s", sig);
		return FALSE;
	}

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_IR,
						DEVICED_INTERFACE_IR,
						method,
						var);
	if (!msg) {
		_E("fail (%s): no reply", method);
		return ret;
	}

	if (!g_variant_get_safe(msg, "(i)", &val))
		_E("fail (%s): no message", method);
	else {
		if ((val == -ENOTSUP) || (val == -ENOSYS)) {
			_I("Not supported feature! (%s): %d", method, val);
			ret = TRUE;
		} else if (val == -ENODEV) {
			_E("fail (%s): device open fail (%d)", method, val);
		} else if (val < 0) {
			_E("fail (%s): returned fail (%d)", method, val);
		} else {
			_I("success (%s): %d", method, val);
			ret = TRUE;
		}
	}

	g_variant_unref(msg);
	return ret;
}

static bool get_ir_isavailable()
{
	return request_ir_method(METHOD_IR_ISAVAILABLE, NULL, NULL);
}

static bool set_ir_transmit(int carrier_frequency, int *pattern, int size)
{
	int freq_pattern[size + 2];

	freq_pattern[0] = carrier_frequency;
	for (int i = 1; i <= size; i++)
		freq_pattern[i] = pattern[i-1];
	freq_pattern[size + 1] = 0;

	return request_ir_method(METHOD_IR_TRANSMIT, "ai", freq_pattern);
}

/*legacy command*/
static bool set_ir_command(char *command)
{
	GVariant *msg;
	int val;
	bool ret = FALSE;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_LED,
						DEVICED_INTERFACE_LED,
						METHOD_IR_SETIRCOMMAND,
						g_variant_new("(s)", command));
	if (!msg) {
		_E("fail : no reply");
		return ret;
	}

	if (!g_variant_get_safe(msg, "(i)", &val))
		_E("fail : no message");
	else {
		if ((val == -ENOTSUP) || (val == -ENOSYS)) {
			_I("Not supported feature! : %d", val);
			ret = TRUE;
		} else if (val < 0) {
			_E("fail : returned fail (%d)", val);
		} else {
			_I("success : %d", val);
			ret = TRUE;
		}
	}

	g_variant_unref(msg);
	return ret;
}

void ir_test_all(int *success, int *fail)
{
	int s = 0;
	int f = 0;

	int pattern[5] = {100, 200, 300, 400, 500};

	(get_ir_isavailable())					? s++ : f++;
	(set_ir_transmit(38000, pattern, 5))	? s++ : f++;
	(set_ir_command("N/A"))					? s++ : f++;


	if (NULL != success)    *success = s;
	if (NULL != fail)       *fail = f;
}

static void ir_init(void *data)
{
	int success = 0;
	int fail = 0;

	_I("start test");

	ir_test_all(&success, &fail);

	_I("Total: %d, Success: %d, Fail: %d", success+fail, success, fail);
}

static void ir_exit(void *data)
{
	_I("end test");
}

static int ir_unit(int argc, char **argv)
{
	if (argc < 4) {
		int success = 0;
		int fail = 0;

		_I("start test");
		ir_test_all(&success, &fail);
		_I("Total: %d, Success: %d, Fail: %d", success+fail, success, fail);
	} else if (0 == strcasecmp(argv[3], METHOD_IR_ISAVAILABLE)) {
		get_ir_isavailable();
	//} else if (0 == strcasecmp(argv[3], METHOD_IR_TRANSMIT)) {
	//	set_ir_transmit();
	} else if (0 == strcasecmp(argv[3], METHOD_IR_SETIRCOMMAND)) {
		set_ir_command(argv[4]);
	} else {
		_E("Unknown test case!!!");
	}

	return 0;
}

static const struct test_ops ir_test_ops = {
	.priority = TEST_PRIORITY_NORMAL,
	.name     = "ir",
	.init     = ir_init,
	.exit    = ir_exit,
	.unit    = ir_unit,
};

TEST_OPS_REGISTER(&ir_test_ops)
