/*
 * test
 *
 * Copyright (c) 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "test.h"

#define METHOD_POWEROFF             "PowerOff"
#define METHOD_POWEROFF_WITH_OPTION "PowerOffWithOption"

#define METHOD_LOWPOWER_START       "Start"
#define METHOD_LOWPOWER_STOP        "Stop"
#define METHOD_LOWPOWER_GETSTATE    "GetState"

static bool set_reboot_method(const char *method, GVariant *param)
{
	GVariant *msg;
	int val;
	bool ret = FALSE;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_POWEROFF,
						DEVICED_INTERFACE_POWEROFF,
						method,
						param);
	if (!msg) {
		_E("fail (%s): no reply", method);
		return ret;
	}

	if (!g_variant_get_safe(msg, "(i)", &val))
		_E("fail (%s): no message", method);
	else {
		if ((val == -ENOTSUP) || (val == -ENOSYS)) {
			_I("Not supported feature! (%s): %d", method, val);
			ret = TRUE;
		} else if (val < 0) {
			_E("fail (%s): returned fail (%d)", method, val);
		} else {
			_I("success (%s): %d", method, val);
			ret = TRUE;
		}
	}

	g_variant_unref(msg);
	return ret;
}

static bool set_reboot(char *type)
{
	return set_reboot_method(METHOD_POWEROFF, g_variant_new("(s)", type));
}

static bool set_reboot_with_option(char *type, char *option)
{
	return set_reboot_method(METHOD_POWEROFF_WITH_OPTION, g_variant_new("(ss)", type, option));
}

static bool request_lowpower_method(const char *method, GVariant *param)
{
	GVariant *msg;
	int val;
	bool ret = FALSE;
	int retry = 0;

	while (retry++ < 3) {
		msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
							DEVICED_PATH_LOWPOWER,
							DEVICED_INTERFACE_LOWPOWER,
							method,
							param);
		if (!msg) {
			_E("fail (%s): no reply", method);
			return ret;
		}

		if (!g_variant_get_safe(msg, "(i)", &val))
			_E("fail (%s): no message", method);
		else {
			if (val == -EAGAIN) {
				g_variant_unref(msg);
				_I("wait! will try again! (%s): %d", method, val);
				sleep(3);
				continue;
			} else if ((val == -ENOTSUP) || (val == -ENOSYS)) {
				_I("Not supported feature! (%s): %d", method, val);
				ret = TRUE;
			} else if (val < 0) {
				_E("fail (%s): returned fail (%d)", method, val);
			} else {
				_I("success (%s): %d", method, val);
				ret = TRUE;
			}
		}
		g_variant_unref(msg);
		break;
	}

	return ret;
}

static bool set_lowpower_start()
{
	return request_lowpower_method(METHOD_LOWPOWER_START, NULL);
}

static bool set_lowpower_stop()
{
	return request_lowpower_method(METHOD_LOWPOWER_STOP, NULL);
}

static bool set_lowpower_getstate()
{
	return request_lowpower_method(METHOD_LOWPOWER_GETSTATE, NULL);
}

void lowpower_test_all(int *success, int *fail)
{
	int s = 0;
	int f = 0;

	(set_lowpower_start())		? s++ : f++;
	(set_lowpower_stop())		? s++ : f++;
	(set_lowpower_getstate())	? s++ : f++;

	if (NULL != success)    *success = s;
	if (NULL != fail)       *fail = f;
}

static void power_init(void *data)
{
	int success = 0;
	int fail = 0;

	_I("start test");

	lowpower_test_all(&success, &fail);

	_I("Total: %d, Success: %d, Fail: %d", success+fail, success, fail);
}

static void power_exit(void *data)
{
	_I("end test");
}

static int power_unit(int argc, char **argv)
{
	if (argc < 4) {
		int success = 0;
		int fail = 0;

		_I("start test");
		lowpower_test_all(&success, &fail);
		_I("Total: %d, Success: %d, Fail: %d", success+fail, success, fail);
	} else if (0 == strcmp(argv[3], METHOD_POWEROFF))
		set_reboot(argv[4]);
	else if (0 == strcasecmp(argv[3], METHOD_POWEROFF_WITH_OPTION))
		set_reboot_with_option(argv[4], argv[5]);
	else if (0 == strcasecmp(argv[3], "lowpower")) {
		if (0 == strcasecmp(argv[4], METHOD_LOWPOWER_START))
			set_lowpower_start();
		else if (0 == strcasecmp(argv[4], METHOD_LOWPOWER_STOP))
			set_lowpower_stop();
		else if (0 == strcasecmp(argv[4], METHOD_LOWPOWER_GETSTATE))
			set_lowpower_getstate();
		else
			_E("Unknown test case!!!");
	} else
		_E("Unknown test case!!!");

	return 0;
}

static const struct test_ops power_test_ops = {
	.priority = TEST_PRIORITY_HIGH,
	.name     = "power",
	.init     = power_init,
	.exit    = power_exit,
	.unit    = power_unit,
};

TEST_OPS_REGISTER(&power_test_ops)
