/*
 * deviced
 *
 * Copyright (c) 2020 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file	display-lock.c
 * @brief	Manage each state's lock. Lock blocks state transition by timeout.
 *
 */
#include "core/common.h"
#include "device-interface.h"
#include "poll.h"
#include "display-signal.h"
#include "lock-detector.h"
#include "display-lock.h"
#include "shared/log-macro.h"

#define METHOD_APP_STATUS	"CheckAppStatus"
#define PID_MAX			6
#define LOCK_TIME_WARNING	60	/* 60 seconds */

static struct _backlight_ops *backlight_ops;
static dd_list *cond_head[S_END];
static int trans_condition;

static struct display_config *display_conf;

bool check_lock_state(int state)
{
	dd_list *elem;
	PmLockNode *t;

	DD_LIST_FOREACH(cond_head[state], elem, t) {
		if (t->background == false)
			return true;
	}

	return false;
}

static void refresh_app_cond()
{
	trans_condition = 0;

	if (check_lock_state(S_NORMAL))
		trans_condition |= MASK_NORMAL;
	if (check_lock_state(S_LCDDIM))
		trans_condition |= MASK_DIM;
	if (check_lock_state(S_LCDOFF))
		trans_condition |= MASK_OFF;
}

static void default_pmlock_check_cb(GVariant *var, void *user_data, GError *err)
{
	pid_t pid = 0;
	int ret, detected = 0;
	char *app_id = NULL;
	enum state_t state = (enum state_t) user_data;
	PmLockNode *node;
	time_t now;
	double diff;

	if (!var)
		return;

	if (!g_variant_get_safe(var, "(iis)", &pid, &detected, &app_id)) {
		_E("Failed to get params from gvariant. expected:%s, type:%s", "(iis)", g_variant_get_type_string(var));
		goto out;
	}

	if (is_app(pid) == 1 && !detected)
		goto out;

	node = find_node(state, pid);
	if (!node)
		goto out;

	time(&now);
	diff = difftime(now, node->time);

	CRITICAL_LOG("%s(%d) has held %s lock for a long time(%.0f s).",
			app_id, pid, state_st(state)->name + 2, diff);

	ret = dbus_handle_emit_dbus_signal(NULL,
						DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						"pmlock_expired",
						g_variant_new("(ii)", pid, (int)diff));
	if (ret < 0)
		_E("Failed to send dbus pmlock_expired");

out:
	if (app_id)
		g_free(app_id);
	g_variant_unref(var);
}

static gboolean default_pmlock_check(void *data)
{
	const char *arr[2];
	char chr_pid[PID_MAX];
	PmLockNode *node;
	enum state_t state;
	pid_t pid;
	int ret;

	if (!data) {
		_E("Invalid parameter.");
		return G_SOURCE_REMOVE;
	}

	node = (PmLockNode *) data;
	state = node->state;
	pid = node->pid;

	if (state == S_LCDOFF && backlight_ops->get_lcd_power() == DPMS_ON) {
		_D("Lcd state is PM_LCD_POWER_ON");
		return G_SOURCE_CONTINUE;
	}

	/* Stop checking lock if process had been terminated */
	if (kill(pid, 0) == -1) {
		node = find_node(state, pid);
		del_node(state, node);
		_I("Process %d not found. Stop checking lock.", pid);
		return G_SOURCE_REMOVE;
	}

	snprintf(chr_pid, sizeof(chr_pid), "%d", pid);
	arr[0] = chr_pid;
	switch (state) {
	case S_NORMAL:
		arr[1] = "normal";
		break;
	case S_LCDDIM:
		arr[1] = "lcddim";
		break;
	case S_LCDOFF:
		arr[1] = "lcdoff";
		break;
	default:
		_E("Invalid state.");
		return G_SOURCE_REMOVE;
	}

	ret = dbus_handle_method_async_with_reply(RESOURCED_BUS_NAME,
		RESOURCED_PATH_PROCESS,
		RESOURCED_INTERFACE_PROCESS,
		METHOD_APP_STATUS,
		"is", arr, default_pmlock_check_cb, -1, (void *)(intptr_t)state);
	if (ret < 0)
		_E("Failed to call dbus method");

	return G_SOURCE_CONTINUE;
}


PmLockNode *find_node(enum state_t s_index, pid_t pid)
{
	dd_list *elem;
	PmLockNode *t = NULL;

	DD_LIST_FOREACH(cond_head[s_index], elem, t) {
		if (t->pid == pid)
			return t;
	}

	return NULL;
}

PmLockNode *add_node(enum state_t s_index, pid_t pid, guint timeout_id,
		bool holdkey_block)
{
	PmLockNode *n;
	time_t now;

	assert(display_conf);

	n = (PmLockNode *) calloc(1, sizeof(PmLockNode));
	if (n == NULL) {
		_E("Not enough memory, failed to alloc lock node.");
		return NULL;
	}

	time(&now);
	n->state = s_index;
	n->pid = pid;
	n->timeout_id = timeout_id;
	n->time = now;
	n->holdkey_block = holdkey_block;
	n->background = false;
	n->broadcast_warning = true;

	if (pid < INTERNAL_LOCK_BASE) {
		/* check if this lock node needs custom-defined lock checker.
		 * n->warning_id would be 0 if fails to register the checker,
		 * or there is no need to use that checker */
		if (display_conf->pmlock_check)
			n->warning_id = display_conf->pmlock_check(n);

		/* use default lock checker */
		if (!n->warning_id)
			n->warning_id = g_timeout_add_seconds(display_conf->lockcheck_timeout,
				default_pmlock_check, (gpointer)n);
	}

	DD_LIST_APPEND(cond_head[s_index], n);

	refresh_app_cond();
	return n;
}

int del_node(enum state_t s_index, PmLockNode *n)
{
	if (n == NULL)
		return 0;

	DD_LIST_REMOVE(cond_head[s_index], n);

	/* delete timer */
	if (n->timeout_id) {
		g_source_remove(n->timeout_id);
		n->timeout_id = 0;
	}
	if (n->warning_id) {
		g_source_remove(n->warning_id);
		n->warning_id = 0;
	}
	if (n->warning_param) {
		g_variant_unref(n->warning_param);
		n->warning_param = NULL;
	}

	free(n);
	refresh_app_cond();
	return 0;
}

int check_lock_condition(enum state_t state)
{
	dd_list *elem;
	PmLockNode *t = NULL;
	int ret = false;
	pid_t owner = getpid();

	_D("check holdkey block : state of %s", state_st(state)->name);

	DD_LIST_FOREACH(cond_head[state], elem, t) {
		if (t->pid != owner && t->background == false) {
			ret = true;
			_I("state change was blocked by pid(%d)!", t->pid);
			break;
		}
	}

	return ret;
}

int check_holdkey_block(enum state_t state)
{
	dd_list *elem;
	PmLockNode *t = NULL;
	int ret = 0;

	_I("Check holdkey block: state of %s", state_st(state)->name);

	if (custom_holdkey_block == true) {
		_I("Custom hold key blocked");
		return 1;
	}

	DD_LIST_FOREACH(cond_head[state], elem, t) {
		if (t->holdkey_block == true) {
			ret = 1;
			_I("Hold key blocked by pid(%d).", t->pid);
			break;
		}
	}

	return ret;
}

int delete_condition(enum state_t state)
{
	dd_list *elem, *next;
	PmLockNode *t = NULL;

	_I("delete condition : state of %s", state_st(state)->name);

	if (!cond_head[state])
		return 0;

	DD_LIST_FOREACH_SAFE(cond_head[state], elem, next, t) {
		if (t->timeout_id > 0) {
			g_source_remove(t->timeout_id);
			t->timeout_id = 0;
		}
		if (state == S_LCDOFF)
			set_process_active(false, t->pid);
		_I("delete node of pid(%d)", t->pid);
		set_unlock_time(t->pid, state);
		del_node(state, t);
	}

	DD_LIST_FREE_LIST(cond_head[state]);
	cond_head[state] = NULL;

	return 0;
}

void print_node(int next)
{
	int ret;
	dd_list *elem;
	PmLockNode *n;
	char buf[30];
	time_t now;
	double diff;

	if (next <= S_START || next >= S_END)
		return;

	time(&now);
	DD_LIST_FOREACH(cond_head[next], elem, n) {
		diff = difftime(now, n->time);
		ctime_r(&n->time, buf);
		buf[strlen(buf) - 1] = 0;

		if (diff > LOCK_TIME_WARNING) {
			if (diff > LOCK_TIME_WARNING * 60 && n->pid < INTERNAL_LOCK_BASE && n->broadcast_warning) {
				ret = dbus_handle_emit_dbus_signal(NULL,
										DEVICED_PATH_DISPLAY,
										DEVICED_INTERFACE_DISPLAY,
										"pmlock_over",
										g_variant_new("(i)", n->pid));
				if (ret < 0)
					_E("Failed to send dbus signal pmlock_over.");
                                n->broadcast_warning = false;
			}
			_W("Over(%.0f s) pid( %5d) lock time(%s)", diff, n->pid, buf);
		} else
			_I("Pid(%5d) lock time(%s)", n->pid, buf);
	}
}

void makeup_trans_condition(void)
{
	check_processes(S_NORMAL);
	check_processes(S_LCDDIM);
	check_processes(S_LCDOFF);
	refresh_app_cond();
}

/* If a lock node in the given state had been dead, delete it form lock list,
 * and returns 1, else returns 0 */
int check_processes(enum state_t prohibit_state)
{
	dd_list *elem, *next;
	PmLockNode *t = NULL;
	int ret = 0;

	DD_LIST_FOREACH_SAFE(cond_head[prohibit_state], elem, next, t) {
		if (t->pid == -1 || (t->pid < INTERNAL_LOCK_BASE && kill(t->pid, 0) == -1)) {
			_E("%d process does not exist, delete the REQ"
				" - prohibit state %d ",
				t->pid, prohibit_state);
			ret = 1;
			set_unlock_time(t->pid, prohibit_state);
			del_node(prohibit_state, t);
		}
	}

	return ret;
}

int get_trans_condition(void)
{
	return trans_condition;
}

dd_list *get_cond_head(enum state_t s_index)
{
	return cond_head[s_index];
}

int process_background(void *data)
{
	pid_t pid;
	PmLockNode *node;

	pid = *(pid_t *)data;

	node = find_node(S_NORMAL, pid);
	if (node) {
		node->background = true;
		_I("Process(%d) is background, then PM will be unlocked LCD_NORMAL.", pid);
	}

	return 0;
}

int process_foreground(void *data)
{
	pid_t pid;
	PmLockNode *node;

	pid = *(pid_t *)data;

	node = find_node(S_NORMAL, pid);
	if (node) {
		node->background = false;
		_I("Process(%d) is foreground, then PM will be locked LCD_NORMAL.", pid);
	}

	return 0;
}

/* update transition condition for application requrements */
void update_lock_timer(PMMsg *data,
		PmLockNode *node, guint timeout_id)
{
	time_t now;

	if (data->timeout > 0) {
		time(&now);
		node->time = now;
	}

	if (node->timeout_id)
		g_source_remove(node->timeout_id);
	node->timeout_id = timeout_id;
}

static void __CONSTRUCTOR__ initialize(void)
{
	backlight_ops = get_var_backlight_ops();
	if (!backlight_ops)
		_E("Failed to get backlight operator variable.");

	display_conf = get_var_display_config();
	if (!display_conf) {
		_E("Failed to get display config variable.");
	}
}
