/*
 * deviced
 *
 * Copyright (c) 2020 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "display-signal.h"
#include "poll.h"
#include "shared/plugin.h"

#define ACTIVE_ACT "active"
#define INACTIVE_ACT "inactive"

static const char *lcdon_sig_lookup[SIGNAL_MAX] = {
	[SIGNAL_PRE]  = "LCDOn",
	[SIGNAL_POST] = "LCDOnCompleted",
};
static const char *lcdoff_sig_lookup[SIGNAL_MAX] = {
	[SIGNAL_PRE]  = "LCDOff",
	[SIGNAL_POST] = "LCDOffCompleted",
};

static struct display_plugin *disp_plgn;
static int pmstate_suspend = PM_SUSPEND;
static long displayoff_time;

void broadcast_pm_suspend(void)
{
	int ret;
	long now;

	if (pmstate_suspend)
		return;

	_D("PM will be changed to sleep.");

	pmstate_suspend = PM_SUSPEND;
	now = clock_gettime_to_long();
	ret = dbus_handle_emit_dbus_signal(NULL,
							DEVICED_PATH_DISPLAY,
							DEVICED_INTERFACE_DISPLAY,
							"sleep",
							g_variant_new("(t)", (guint64)now));
	if (ret < 0)
		_E("Failed to send dbus signal sleep.");
}

void broadcast_pm_wakeup(void)
{
	int ret;
	long now;

	if (!pmstate_suspend)
		return;

	 _D("PM is changed to wakeup.");

	 pmstate_suspend = PM_WAKEUP;
	 now = clock_gettime_to_long();
	 ret = dbus_handle_emit_dbus_signal(NULL,
							 DEVICED_PATH_DISPLAY,
							 DEVICED_INTERFACE_DISPLAY,
							 "wakeup",
							 g_variant_new("(t)", (guint64)now));
	 if (ret < 0)
		 _E("Failed to send dbus signal wakeup.");

}

void broadcast_lcd_on(enum signal_type type, enum device_flags flags)
{
	const char *str;
	const char *signal;
	int ret;
	long diff = 0;

	if (type <= SIGNAL_INVALID || type >= SIGNAL_MAX) {
		_E("Invalid signal type(%d).", type);
		return;
	}

	if (type == SIGNAL_PRE && displayoff_time != 0)
		diff = clock_gettime_to_long() - displayoff_time;

	if (!disp_plgn->device_flags_to_string) {
		_E("Cannot convert device_flags to string.");
		return;
	}

	str = disp_plgn->device_flags_to_string(flags);

	signal = lcdon_sig_lookup[type];
	_I("lcdstep : Broadcast signal(%s:%s).", signal, str);
	ret = dbus_handle_emit_dbus_signal_sync(NULL,
						DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						signal,
						g_variant_new("(si)", str, diff));
	if (ret < 0)
		_E("Failed to send dbus signal(%s)", signal);
}

void broadcast_lcd_off(enum signal_type type, enum device_flags flags)
{
	const char *str;
	const char *signal;
	int ret;

	if (type <= SIGNAL_INVALID || type >= SIGNAL_MAX) {
		_E("Invalid signal type(%d).", type);
		return;
	}

	if (type == SIGNAL_PRE)
		displayoff_time = clock_gettime_to_long();

	signal = lcdoff_sig_lookup[type];

	if (!disp_plgn->device_flags_to_string) {
		_E("Cannot convert device_flags to string.");
		return;
	}

	str = disp_plgn->device_flags_to_string(flags);

	_I("lcdstep : Broadcast signal(%s).", signal);
	ret = dbus_handle_emit_dbus_signal_sync(NULL,
						DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						signal,
						g_variant_new("(s)", str));
	if (ret < 0)
		_E("Failed to send dbus signal(%s)", signal);
}

void broadcast_lcd_off_late(enum device_flags flags)
{
	static enum device_flags late_flags;

	if (flags & LCD_OFF_LATE_MODE) {
		broadcast_lcd_off(SIGNAL_POST, late_flags);
		device_notify(DEVICE_NOTIFIER_LCD_OFF_COMPLETE, NULL);
	} else {
		late_flags = flags;
	}
}

void set_process_active(bool flag, pid_t pid)
{
	int ret;

	if (pid >= INTERNAL_LOCK_BASE)
		return;

	/* Send dbus signal to resourced */
	ret = dbus_handle_emit_dbus_signal(NULL,
							RESOURCED_PATH_PROCESS,
						    RESOURCED_INTERFACE_PROCESS,
						    RESOURCED_METHOD_ACTIVE,
						    g_variant_new("(si)", (flag ? ACTIVE_ACT : INACTIVE_ACT), pid));
	if (ret < 0)
		_E("Failed to send dbus signal to resourced.");
}

static void __CONSTRUCTOR__ initialize(void)
{
	disp_plgn = get_var_display_plugin();
	if (!disp_plgn)
		_E("Failed to get display plugin variable.");
}
