/*
 * deviced
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>

#include "util.h"
#include "core.h"
#include "display-ops.h"
#include "dd-display.h"
#include "core/list.h"
#include "core/common.h"
#include "core/config-parser.h"

static int pm_cur_state;
static int pm_old_state;
static unsigned int pm_status_flag;

inline int get_pm_cur_state(void)
{
	return pm_cur_state;
}

inline void set_pm_cur_state(int cur_state)
{
	pm_cur_state = cur_state;
}

inline int get_pm_old_state(void)
{
	return pm_old_state;
}

inline void set_pm_old_state(int old_state)
{
	pm_old_state = old_state;
}

inline unsigned int get_pm_status_flag(void)
{
	return pm_status_flag;
}

inline void set_pm_status_flag(unsigned int status_flag)
{
	pm_status_flag |= status_flag;
}

inline void clear_pm_status_flag(unsigned int status_flag)
{
	pm_status_flag &= ~status_flag;
}

void lcd_direct_control(enum dpms_state state, int flags)
{
	const struct device_ops *ops = NULL;
	dd_list *l = NULL;

	switch (state) {
	case DPMS_ON:
		DD_LIST_FOREACH(lcdon_ops, l, ops)
			ops->start(flags);
		break;
	case DPMS_OFF:
		DD_LIST_FOREACH(lcdon_ops, l, ops)
			ops->stop(flags);
		break;
	default:
		_E("state is wrong value %d", state);
		break;
	}
}

int display_load_config(struct parse_result *result, void *user_data)
{
	struct display_config *c = user_data;

	_D("display_load_config: section=%s name=%s value=%s", result->section, result->name, result->value);

	if (!c)
		return -EINVAL;

	if (!MATCH(result->section, "Display"))
		return 0;

	if (MATCH(result->name, "LockScreenWaitingTime")) {
		SET_CONF(c->lock_wait_time, atof(result->value));
		_D("lock wait time is %.3f", c->lock_wait_time);
	} else if (MATCH(result->name, "LongPressInterval")) {
		SET_CONF(c->longpress_interval, atof(result->value));
		_D("long press interval is %.3f", c->longpress_interval);
	} else if (MATCH(result->name, "LightSensorSamplingInterval")) {
		SET_CONF(c->lightsensor_interval, atof(result->value));
		_D("lightsensor interval is %.3f", c->lightsensor_interval);
	} else if (MATCH(result->name, "LCDOffTimeout")) {
		SET_CONF(c->lcdoff_timeout, atoi(result->value));
		_D("lcdoff timeout is %d ms", c->lcdoff_timeout);
	} else if (MATCH(result->name, "BrightnessChangeStep")) {
		SET_CONF(c->brightness_change_step, atoi(result->value));
		_D("brightness change step is %d", c->brightness_change_step);
	} else if (MATCH(result->name, "LCDAlwaysOn")) {
		c->lcd_always_on = (MATCH(result->value, "yes") ? 1 : 0);
		_D("LCD always on is %d", c->lcd_always_on);
	} else if (MATCH(result->name, "Dimming")) {
		c->dimming = (MATCH(result->value, "yes") ? 1 : 0);
		 _D("Dimming is %d", c->dimming);
	} else if (MATCH(result->name, "ChangedFrameRateAllowed")) {
		if (strstr(result->value, "setting")) {
			c->framerate_app[REFRESH_SETTING] = 1;
			_D("framerate app is Setting");
		}
		if (strstr(result->value, "all")) {
			memset(c->framerate_app, 1, sizeof(c->framerate_app));
			_D("framerate app is All");
		}
	} else if (MATCH(result->name, "ControlDisplay")) {
		c->control_display = (MATCH(result->value, "yes") ? 1 : 0);
		_D("ControlDisplay is %d", c->control_display);
	} else if (MATCH(result->name, "PowerKeyDoublePressSupport")) {
		c->powerkey_doublepress = (MATCH(result->value, "yes") ? 1 : 0);
		_D("PowerKeyDoublePressSupport is %d", c->powerkey_doublepress);
	} else if (MATCH(result->name, "AccelSensorOn")) {
		c->accel_sensor_on = (MATCH(result->value, "yes") ? 1 : 0);
		_D("AccelSensorOn is %d", c->accel_sensor_on);
	} else if (MATCH(result->name, "ContinuousSampling")) {
		c->continuous_sampling = (MATCH(result->value, "yes") ? 1 : 0);
		_D("ContinuousSampling is %d", c->continuous_sampling);
	} else if (MATCH(result->name, "TimeoutEnable")) {
		c->timeout_enable = (MATCH(result->value, "yes") ? true : false);
		_D("Timeout is %s", c->timeout_enable ? "enalbed" : "disabled");
	} else if (MATCH(result->name, "InputSupport")) {
		c->input_support = (MATCH(result->value, "yes") ? true : false);
		_D("Input is %s", c->input_support ? "supported" : "NOT supported");
	} else if (MATCH(result->name, "LockCheckTimeout")) {
		SET_CONF(c->lockcheck_timeout, atoi(result->value));
		_D("LockCheckTimeout is %d", c->lockcheck_timeout);
	} else if (MATCH(result->name, "AODTSP")) {
		c->aod_tsp = (MATCH(result->value, "yes") ? true : false);
		_D("TSP control at is %d at aod", c->aod_tsp);
	} else if (MATCH(result->name, "SleepSupport")) {
		c->sleep_support = (MATCH(result->value, "yes") ? true : false);
		_D("SleepSupport is %d", c->sleep_support);
	}

	return 0;
}
