/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <linux/input.h>
#include <libinput.h>

#include "util.h"
#include "core.h"
#include "poll.h"
#include "ambient-mode.h"

#define SEAT_NAME   "seat0"

static struct udev *udev;
static struct libinput *li;
static guint efd;

int (*pm_callback) (int, PMMsg *);

static inline void process_event(struct libinput_event *ev)
{
	static const struct device_ops *display_device_ops;
	struct input_event input;
	struct libinput *li;
	struct libinput_event_keyboard *k;
	unsigned int time;
	int fd = 0;
	struct display_config *display_conf = get_var_display_config();
	if(!display_conf) {
		_E("Failed to get display configuration variable.");
		return;
	}

	if (!pm_callback)
		return;

	if (!display_device_ops) {
		display_device_ops = find_device("display");
		if (!display_device_ops)
			return;
	}

	/* do not operate when display stops */
	if (device_get_status(display_device_ops)
			!= DEVICE_OPS_STATUS_START) {
		_E("display status is stop");
		return;
	}

	switch (libinput_event_get_type(ev)) {
	case LIBINPUT_EVENT_KEYBOARD_KEY:
		k = libinput_event_get_keyboard_event(ev);
		time = libinput_event_keyboard_get_time(k);
		li = libinput_event_get_context(ev);

		input.time.tv_sec = MSEC_TO_SEC(time);
		input.time.tv_usec = MSEC_TO_USEC(time % 1000);
		input.type = EV_KEY;
		input.code = libinput_event_keyboard_get_key(k);
		input.value = libinput_event_keyboard_get_key_state(k);

		fd = libinput_get_fd(li);
		_D("time %ld.%06ld type %d code %d value %d fd %d",
				input.time.tv_sec, input.time.tv_usec, input.type,
				input.code, input.value, fd);
		break;
	case LIBINPUT_EVENT_POINTER_MOTION:
	case LIBINPUT_EVENT_POINTER_BUTTON:
		input.type = EV_REL;
		break;
	case LIBINPUT_EVENT_TOUCH_DOWN:
		input.type = EV_ABS;
		input.value = KEY_PRESSED;
		break;
	case LIBINPUT_EVENT_TOUCH_UP:
		input.type = EV_ABS;
		input.value = KEY_RELEASED;
		break;
	case LIBINPUT_EVENT_TOUCH_MOTION:
	case LIBINPUT_EVENT_TOUCH_FRAME:
		if (touch_event_blocked() && !ambient_get_state() && !display_conf->touch_wakeup)
			return;
		input.type = EV_ABS;
		input.value = KEY_BEING_PRESSED;
		break;
	default:
		return;
	}

	if (CHECK_OPS(keyfilter_ops, check) &&
	    keyfilter_ops->check(&input, fd) != 0)
		return;

	/* lcd on or update lcd timeout */
	(*pm_callback) (INPUT_POLL_EVENT, NULL);
}

static gboolean input_handler(gint fd, GIOCondition cond, void *data)
{
	struct libinput_event *ev;
	struct libinput *input = (struct libinput *)data;

	if (!input)
		return G_SOURCE_CONTINUE;

	libinput_dispatch(input);

	while ((ev = libinput_get_event(input))) {
		process_event(ev);

		libinput_event_destroy(ev);
		libinput_dispatch(input);
	}

	return G_SOURCE_CONTINUE;
}

static int open_restricted(const char *path, int flags, void *user_data)
{
	int fd;
	unsigned int clockid = CLOCK_MONOTONIC;

	if (!path)
		return -EINVAL;

	fd = open(path, flags);
	if (fd >= 0) {
		/* TODO Why does fd change the clock? */
		if (ioctl(fd, EVIOCSCLOCKID, &clockid) < 0)
			_E("fail to change clock %s", path);
	}

	return fd < 0 ? -errno : fd;
}

static void close_restricted(int fd, void *user_data)
{
	close(fd);
}

static const struct libinput_interface interface = {
	.open_restricted = open_restricted,
	.close_restricted = close_restricted,
};

gboolean init_input(gpointer data)
{
	int ret;
	int fd;

	udev = udev_new();
	if (!udev) {
		_E("fail to create udev library context");
		return G_SOURCE_REMOVE;
	}

	li = libinput_udev_create_context(&interface, NULL, udev);
	if (!li) {
		_E("fail to create a new libinput context from udev");
		return G_SOURCE_REMOVE;
	}

	ret = libinput_udev_assign_seat(li, SEAT_NAME);
	if (ret < 0) {
		_E("fail to assign a seat");
		return G_SOURCE_REMOVE;
	}

	fd = libinput_get_fd(li);
	if (fd < 0) {
		_E("fail to get file descriptor from libinput context");
		return G_SOURCE_REMOVE;
	}

	/* add to poll handler */
	efd = g_unix_fd_add(fd, G_IO_IN,
			input_handler,
			(void *)((intptr_t)li));
	if (!efd) {
		_E("fail to g_unix_fd_add");
		/* TODO Does it really need close()? */
		close(fd);
		return G_SOURCE_REMOVE;
	}

	return G_SOURCE_REMOVE;
}

int exit_input(void)
{
	if (efd) {
		g_source_remove(efd);
		efd = -1;
	}
	/* todo: close fd */

	if (li)
		libinput_unref(li);

	if (udev)
		udev_unref(udev);

	return 0;
}
