/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <vconf.h>
#include <libsyscommon/dbus-system.h>
#include "core/log.h"
#include "core/list.h"
#include "core/device-notifier.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "extcon.h"
#include "shared/plugin.h"

#define METHOD_GET_HDMI		"GetHDMI"
#define SIGNAL_HDMI_STATE	"ChangedHDMI"

#define HDMI_NUMBER_MAX	(64)					// max number of supported hdmi
#define HDMI_BITMAP_T	(sizeof(hdmi_bitmap_t) * BITS_PER_CHAR)	// size of hdmi_bitmap_t

/* macro to set/unset bit */
#define HDMI_SET_BIT(num)	(hdmi_bitmap |= (1ULL << ((num) % HDMI_BITMAP_T)))
#define HDMI_UNSET_BIT(num)	(hdmi_bitmap &= ~(1ULL << ((num) % HDMI_BITMAP_T)))

typedef unsigned long long hdmi_bitmap_t;

static hdmi_bitmap_t hdmi_bitmap;
static struct display_plugin *disp_plgn;
static struct extcon_ops hdmi_extcon_ops;

static void hdmi_send_broadcast(int status)
{
	static int old;
	int ret;

	if (old == status)
		return;

	_I("Broadcast hdmi status(%d)", status);
	old = status;

	ret = dbus_handle_emit_dbus_signal(NULL,
						DEVICED_PATH_SYSNOTI,
						DEVICED_INTERFACE_SYSNOTI,
						SIGNAL_HDMI_STATE,
						g_variant_new("(i)", status));
	if (ret < 0)
		_E("Failed to send dbus signal(%s)", SIGNAL_HDMI_STATE);
}

/* if hdmi_bitmap is bigger than 0, we assume hdmi is connected */
static int get_hdmi_status(void)
{
	if (hdmi_bitmap > 0)
		return HDMI_CONNECTED;

	return HDMI_DISCONNECTED;
}


static void update_hdmi_bitmap(int index, int status)
{
	if (status == HDMI_CONNECTED)
		HDMI_SET_BIT(index);
	else
		HDMI_UNSET_BIT(index);
}

static void update_all_hdmi_bitmap(int status)
{
	if (status == HDMI_CONNECTED)
		hdmi_bitmap = ULLONG_MAX;
	else
		hdmi_bitmap = 0;
}

static int hdmi_update(const char *index, int status)
{
	int ret;
	int num = -1;
	int hdmi_status;

	_I("Hdmi changed");
	if (disp_plgn->pm_change_internal)
		disp_plgn->pm_change_internal(INTERNAL_LOCK_HDMI, LCD_NORMAL);

	if (!index)
		update_all_hdmi_bitmap(status);
	else {
		sscanf(index, "extcon%d", &num);
		if (num < 0 || num >= HDMI_NUMBER_MAX) {
			_E("Invalid HDMI number: %d", num);
			return -EINVAL;
		}
		update_hdmi_bitmap(num, status);
	}

	hdmi_status = get_hdmi_status();
	ret = vconf_set_int(VCONFKEY_SYSMAN_HDMI, hdmi_status);
	if (ret < 0)
		_E("Failed to set vconf value for hdmi: %d", vconf_get_ext_errno());

	hdmi_extcon_ops.status = hdmi_status;
	hdmi_send_broadcast(hdmi_status);

	if (hdmi_status == HDMI_CONNECTED) {
		if (disp_plgn->pm_lock_internal)
			disp_plgn->pm_lock_internal(INTERNAL_LOCK_HDMI, LCD_DIM, STAY_CUR_STATE, 0);
	} else {
		if (disp_plgn->pm_unlock_internal)
			disp_plgn->pm_unlock_internal(INTERNAL_LOCK_HDMI, LCD_DIM, PM_SLEEP_MARGIN);
	}

	return 0;
}

static int display_changed(void *data)
{
	enum state_t state;
	int hdmi;

	if (!data)
		return 0;

	state = *(int *)data;
	if (state != S_NORMAL)
		return 0;

	hdmi = hdmi_extcon_ops.status;
	if (hdmi == 0) {
		if (disp_plgn->pm_lock_internal)
			disp_plgn->pm_lock_internal(INTERNAL_LOCK_HDMI, LCD_DIM, STAY_CUR_STATE, 0);
		_I("Hdmi is connected. Dim lock is on.");
	}
	return 0;
}

static GVariant *dbus_hdmi_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return g_variant_new("(i)", hdmi_extcon_ops.status);
}

static const dbus_method_s dbus_methods[] = {
	{ METHOD_GET_HDMI,       NULL, "i", dbus_hdmi_handler },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_SYSNOTI,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static void hdmi_init(void *data)
{
	int ret;

	register_notifier(DEVICE_NOTIFIER_LCD, display_changed);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_SYSNOTI, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);
}

static void hdmi_exit(void *data)
{
	unregister_notifier(DEVICE_NOTIFIER_LCD, display_changed);
}

static struct extcon_ops hdmi_extcon_ops = {
	.name   = EXTCON_CABLE_HDMI,
	.init   = hdmi_init,
	.exit   = hdmi_exit,
	.update = hdmi_update,
};

EXTCON_OPS_REGISTER(hdmi_extcon_ops)

static void __CONSTRUCTOR__ initialize(void)
{
	disp_plgn = get_var_display_plugin();
	if (!disp_plgn)
		_E("Failed to get display plugin variable.");
}
