/*
 * deviced
 *
 * Copyright (c) 2020 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <vconf.h>

#include "doze.h"
#include "core/log.h"
#include "core/device-notifier.h"
#include "display/core.h"

#define DOZE_DISABLED 0
#define DOZE_ENABLED  1
static int doze_mode; /* doze from setting */

enum doze_level {
	DOZE_LEVEL_0 = 0, /* doze off mode */
	DOZE_LEVEL_1,
	DOZE_LEVEL_MAX,
};
static enum doze_level doze_level;
static int display_state;
static bool doze_initialized = false;

int doze_pmqos_request(int req)
{
	int ret = 0;

	ret = dbus_handle_method_async_var(PASS_BUS_NAME,
					PASS_PATH_PMQOS,
					PASS_INTERFACE_PMQOS,
					PASS_METHOD_DOZE,
					g_variant_new("(si)", "Doze", req));
	if (ret < 0)
		_E("Failed to request pmqos: %d", ret);

	return ret;
}

static int change_doze_level(enum doze_level level)
{
	int ret;

	if (level < DOZE_LEVEL_0 || level >= DOZE_LEVEL_MAX) {
		_E("Invalid doze level %d.", level);
		return -EINVAL;
	}

	if (doze_level == level)
		return 0;

	_D("Change doze level, %d -> %d.", doze_level, level);
	doze_level = level;

	ret = vconf_set_int(VCONFKEY_SYSMAN_DOZE_MODE_LEVEL, doze_level);
	if (ret < 0) {
		_E("Failed to set doze level, %d.", vconf_get_ext_errno());
		return ret;
	}

	return 0;
}

static void _enter_doze(void)
{
	/* already in doze mode */
	if (doze_level >= DOZE_LEVEL_1)
		return;

	_D("Entering doze mode.");
	change_doze_level(DOZE_LEVEL_1);
	doze_pmqos_request(DOZE_PMQOS_START);
}

static void _leave_doze(void)
{
	/* already not in doze mode */
	if (doze_level == DOZE_LEVEL_0)
		return;

	_D("Leaving doze mode.");
	change_doze_level(DOZE_LEVEL_0);
	doze_pmqos_request(DOZE_PMQOS_STOP);
}

void enter_doze(void)
{
	if (!doze_initialized)
		return;

	if (doze_mode == DOZE_DISABLED)
		return;

	_enter_doze();
}

void leave_doze(void)
{
	if (!doze_initialized)
		return;

	if (doze_mode == DOZE_DISABLED)
		return;

	_leave_doze();
}

static int display_changed_cb(void *data)
{
	display_state = *(int *)data;

	return 0;
}

static void setting_doze_mode_cb(keynode_t *key, void *data)
{
	int val;

	val = vconf_keynode_get_bool(key);

	if (val < 0) {
		_E("Failed to get vconf value for doze, %d.", vconf_get_ext_errno());
		return;
	}

	if (doze_mode == val)
		return;

	doze_mode = val;

	if (!doze_initialized)
		return;

	if (doze_mode == DOZE_DISABLED) {
		_leave_doze();
		_D("Doze mode disabled.");
	} else if (doze_mode == DOZE_ENABLED) {
		_D("Doze mode enabled.");
		if (display_state == S_LCDOFF || display_state == S_SLEEP)
			_enter_doze();
	}

}

int doze_init(void)
{
	int ret;

	ret = vconf_get_bool(VCONFKEY_SETAPPL_DOZE_MODE, &doze_mode);
	if (ret < 0) {
		_E("Failed to get vconf value for doze, %d.", vconf_get_ext_errno());
		return ret;
	}

	ret = vconf_notify_key_changed(VCONFKEY_SETAPPL_DOZE_MODE, setting_doze_mode_cb, NULL);
	if (ret < 0) {
		_E("Failed to register vconf notify.");
		return ret;
	}

	doze_level = DOZE_LEVEL_0;

	register_notifier(DEVICE_NOTIFIER_LCD, display_changed_cb);

	doze_initialized = true;
	_D("Doze mode initialized. doze_mode: %d.", doze_mode);

	return 0;
}

int doze_exit(void)
{
	doze_initialized = false;

	unregister_notifier(DEVICE_NOTIFIER_LCD, display_changed_cb);

	return 0;
}
