/*
 * deviced
 *
 * Copyright (c) 2020 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include <errno.h>
#include <dlfcn.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"
#include "core/devices.h"
#include "core/common.h"
#include "core/device-notifier.h"
#include "vconf.h"
#include "device-node.h"
#include "display/display-dpms.h"
#include "display/display.h"
#include "power/boot.h"
#include "power-control.h"

static int mainlock_status = POWER_UNLOCK;
static int vital_service;
static bool vital_sleep;
static int vital_support = -2;

#ifdef ENABLE_PM_LOG

typedef struct _pm_history {
	time_t time;
	enum pm_log_type log_type;
	int keycode;
} pm_history;

static int max_history_count = MAX_LOG_COUNT;
static pm_history pm_history_log[MAX_LOG_COUNT] = {{0, }, };
static int history_count = 0;

static const char history_string[PM_LOG_MAX][15] = {
	"PRESS", "LONG PRESS", "RELEASE", "LCD ON", "LCD ON COMPL", "LCD ON FAIL",
	"LCD DIM", "LCD DIM FAIL", "LCD OFF", "LCD OFF COMPL", "LCD OFF FAIL",
	"LCD FAIL", "SLEEP"};

void pm_history_init()
{
	memset(pm_history_log, 0x0, sizeof(pm_history_log));
	history_count = 0;
	max_history_count = MAX_LOG_COUNT;
}

void pm_history_save(enum pm_log_type log_type, int code)
{
	time_t now;

	time(&now);
	pm_history_log[history_count].time = now;
	pm_history_log[history_count].log_type = log_type;
	pm_history_log[history_count].keycode = code;
	history_count++;

	if (history_count >= max_history_count)
		history_count = 0;
}

void pm_history_print(int fd, int count)
{
	int start_index, index, i;
	int ret;
	char buf[255];
	char time_buf[30];

	if (count <= 0 || count > max_history_count)
		return;

	start_index = (history_count - count + max_history_count)
		    % max_history_count;

	for (i = 0; i < count; i++) {
		index = (start_index + i) % max_history_count;

		if (pm_history_log[index].time == 0)
			continue;

		if (pm_history_log[index].log_type >= PM_LOG_MAX)
			continue;
		ctime_r(&pm_history_log[index].time, time_buf);
		time_buf[strlen(time_buf) - 1] = 0;
		snprintf(buf, sizeof(buf), "[%3d] %15s %3d %s\n",
			index,
			history_string[pm_history_log[index].log_type],
			pm_history_log[index].keycode,
			time_buf);
		ret = write(fd, buf, strlen(buf));
		if (ret < 0)
			_E("Write() failed: %d", errno);
	}
}
#endif

bool vital_mode(void)
{
	return vital_sleep;
}

static int vital_mode_support(void)
{
	if (vital_support < 0) {
		FILE *fp;

		fp = fopen(FREEZER_VITAL_WAKEUP_CGROUP, "r");
		if (fp == NULL) {
			_E("%s open failed", FREEZER_VITAL_WAKEUP_CGROUP);
			/* read max 2 times to check if this file exist */
			vital_support++;
			return 0;
		}
		vital_support = 1;
		fclose(fp);
	}
	return vital_support;
}

int suspend_other_process(int type)
{
	int ret = 0;
	char buf[8];
	const char *command[1];

	if (vital_service == type)
		return ret;

	if (type == VITAL_WAKEUP && vital_service > VITAL_SLEEP)
		return ret;

	vital_service = type;

	if (!vital_mode_support())
		return ret;

	if (type == VITAL_SLEEP) {
		snprintf(buf, sizeof(buf), "%s", "sleep");
		command[0] = buf;
		dbus_handle_method_sync_timeout(RESOURCED_BUS_NAME,
						RESOURCED_PATH_FREEZER,
						RESOURCED_INTERFACE_FREEZER,
						"SetSuspend",
						"s",
						command,
						SET_SUSPEND_TIME*1000);
		vital_sleep = true;
	} else if (type == VITAL_WAKEUP) {
		snprintf(buf, sizeof(buf), "%s", "wakeup");
		command[0] = buf;
		ret = dbus_handle_method_async(RESOURCED_BUS_NAME,
						RESOURCED_PATH_FREEZER,
						RESOURCED_INTERFACE_FREEZER,
						"SetSuspend",
						"s", command);
	} else if (type == VITAL_EXIT) {
		snprintf(buf, sizeof(buf), "%s", "exit");
		command[0] = buf;
		ret = dbus_handle_method_async(RESOURCED_BUS_NAME,
						RESOURCED_PATH_FREEZER,
						RESOURCED_INTERFACE_FREEZER,
						"SetSuspend",
						"s", command);
		vital_sleep = false;
	}
	return ret;
}

int vital_state_changed(void *data)
{
	int type;

	assert(data);

	type = *(int *)data;
	if (type == VITAL_EXIT)
		suspend_other_process(VITAL_EXIT);

	return 0;
}

int pm_suspend(void)
{
	int ret;

	_I("system suspend");
	ret = sys_set_str(POWER_STATE_PATH, "mem");
	_I("System resume: %d", ret);
	return 0;
}

int pm_enable_autosleep(void)
{
	_I("System autosleep enabled.");
	return sys_set_str(POWER_AUTOSLEEP_PATH, "mem");
}

int pm_power_lock(void)
{
	_I("system power lock");
	suspend_other_process(VITAL_WAKEUP);
	mainlock_status = POWER_LOCK;

	return sys_set_str(POWER_LOCK_PATH, "mainlock");
}

int pm_get_power_lock(void)
{
	return mainlock_status;
}

int pm_get_power_lock_support(void)
{
	static int power_lock_support = -1;
	int ret;

	if (power_lock_support >= 0)
		goto out;

	ret = sys_check_node(POWER_LOCK_PATH);
	if (ret < 0)
		power_lock_support = false;
	else
		power_lock_support = true;

	_I("System power lock: %s",
			(power_lock_support ? "support" : "not support"));

out:
	return power_lock_support;
}

int pm_power_unlock(void)
{
	_I("system power unlock");
	suspend_other_process(VITAL_SLEEP);
	mainlock_status = POWER_UNLOCK;

	return sys_set_str(POWER_UNLOCK_PATH, "mainlock");
}

int check_wakeup_src(void)
{
	/*  TODO if nedded.
	 * return wackeup source. user input or device interrupts? (EVENT_DEVICE or EVENT_INPUT)
	 */
	return EVENT_DEVICE;
}

int get_wakeup_count(int *cnt)
{
	int ret;
	int wakeup_count;

	if (!cnt)
		return -EINVAL;

	ret = sys_get_int(POWER_WAKEUP_PATH, &wakeup_count);
	if (ret < 0)
		return ret;

	*cnt = wakeup_count;
	return 0;
}

int set_wakeup_count(int cnt)
{
	int ret;

	ret = sys_set_int(POWER_WAKEUP_PATH, cnt);
	if (ret < 0)
		return ret;

	return 0;
}
