/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <vconf.h>
#include <hw/usb_client.h>
#include <hw/usb_gadget.h>
#include <libsyscommon/dbus-systemd.h>

#include "core/log.h"
#include "core/udev.h"
#include "display/poll.h"
#include "display/display-ops.h"
#include "shared/plugin.h"

#include "usb.h"
#include "usb-debug.h"
#include "usb-tethering.h"

static int usb_change_gadget(unsigned mode);

static struct display_plugin *disp_plgn;
static struct usb_client *usb_client;
static struct usb_gadget_translator *gadget_translator;

static struct usb_gadget_translator *gadget_translator_probe(void)
{
	struct hw_info *info;
	struct hw_common *common;
	struct usb_gadget_translator *gadget;

	if (hw_get_info(USB_GADGET_DEVICE_ID, (const struct hw_info **)&info)) {
		_E("No USB gadget translator.");
		return NULL;
	}

	if (!info->open) {
		_E("USB gadget does not have open() ");
		return NULL;
	}

	if (info->open(info, NULL, &common) < 0) {
		_E("Failed to open USB gadget translator.");
		return NULL;
	}

	gadget = container_of(common, struct usb_gadget_translator, common);
	if (!gadget->id_to_gadget || !gadget->cleanup_gadget) {
		_E("Invalid USB gadget translator.");

		if (gadget->common.info)
			gadget->common.info->close(&gadget->common);

		return NULL;
	}

	return gadget;
}

static struct usb_client *usb_client_probe(const char *id)
{
	struct hw_info *info;
	struct hw_common *common;
	struct usb_client *client;

	_I("Loading USB client. id=%s", id);

	if (hw_get_info(id, (const struct hw_info **)&info)) {
		_I("No USB client. id=%s", id);
		return NULL;
	}

	if (!info->open) {
		_E("USB client does not have open(). id=%s", id);
		return NULL;
	}

	if (info->open(info, NULL, &common) < 0) {
		_E("Failed to open USB client. id=%s", id);
		return NULL;
	}

	client = container_of(common, struct usb_client, common);
	if (!client->reconfigure_gadget || !client->enable || !client->disable) {
		_E("Invalid USB client. id=%s", id);

		if (client->common.info->close)
			client->common.info->close(&client->common);

		return NULL;
	}

	_I("Success lodging USB client. id=%s", id);
	return client;
}

static int usb_probe(void)
{
	/* Both the gadget and client should be valid. */
	if (gadget_translator || usb_client) {
		if (gadget_translator && usb_client)
			return 0;
		return -EINVAL;
	}

	/* gadget_translator */
	gadget_translator = gadget_translator_probe();
	if (!gadget_translator)
		return -ENODEV;

	/* usb_client */
	usb_client = usb_client_probe(USB_CLIENT_HARDWARE_DEVICE_ID);

	if (!usb_client)
		usb_client = usb_client_probe(USB_CFS_CLIENT_HARDWARE_DEVICE_ID);

	if (!usb_client) {
		if (gadget_translator->common.info)
			gadget_translator->common.info->close(&gadget_translator->common);
		gadget_translator = NULL;

		return -ENODEV;
	}

	return 0;
}

static void usb_release()
{
	if (usb_client) {
		if (usb_client->common.info->close)
			usb_client->common.info->close(&usb_client->common);
		usb_client = NULL;
	}

	if (gadget_translator) {
		if (gadget_translator->common.info->close)
			gadget_translator->common.info->close(
				&gadget_translator->common);
		gadget_translator = NULL;
	}
}

static int usb_config_init(void)
{
	unsigned int mode = usb_state_get_selected_mode();

	return usb_change_gadget(mode);
}

static void usb_config_deinit(void)
{
	if (!usb_client)
		goto no_hal;

	usb_client->disable(usb_client);

	return;

no_hal:
	/* TODO. Maybe some default action here? */
	return;
}

static int usb_config_enable(void)
{
	if (!usb_client)
		goto no_hal;

	return usb_client->enable(usb_client);

no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

static int usb_config_disable(void)
{
	if (!usb_client)
		goto no_hal;

	usb_client->disable(usb_client);

	return 0;

no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

/* Precondition: USB_FUNCTION_NONE */
static int usb_change_gadget(unsigned mode)
{
	int ret;
	struct usb_gadget *gadget;
	struct usb_gadget_id gadget_id;

	if (!gadget_translator || !usb_client) {
		ret = -ENODEV;
		goto out;
	}

	memset(&gadget_id, 0, sizeof(gadget_id));
	gadget_id.function_mask = mode;

	ret = gadget_translator->id_to_gadget(&gadget_id, &gadget);
	if (ret) {
		_E("Unable to translate USB mode id into gadget: %d", ret);
		goto out;
	}

	ret = usb_client->reconfigure_gadget(usb_client, gadget);
	gadget_translator->cleanup_gadget(gadget);
	if (ret) {
		_E("Unable to configure USB gadget: %d", ret);
		goto out;
	}

	_I("USB gadget changed to (%#x)", mode);

	return ret;

out:
	/* Although this function has a USB_FUNCTION_NONE precondition, it is to protect against coding mistakes. */
	(void)usb_state_set_current_mode(USB_FUNCTION_NONE); /* because usb does not work properly */
	return ret;
}

/* Precondition: USB_CONNECTED, USB_FUNCTION_NONE */
static int usb_enable(unsigned int mode)
{
	int ret;

	ret = usb_config_enable();
	if (ret < 0) {
		_E("Failed to enable USB config: %d", ret);
		goto out;
	}

	send_usb_state_changed_event(VCONFKEY_SYSMAN_USB_AVAILABLE);
	(void)usb_state_set_current_mode(mode);
	change_usb_state_notification_handler(mode);

	return 0;

out:
	/* Although this function has a USB_FUNCTION_NONE precondition, it is to protect against coding mistakes. */
	(void)usb_state_set_current_mode(USB_FUNCTION_NONE); /* because usb does not work properly */
	return ret;
}

/* Precondition: N/A */
static int usb_disable(void)
{
	int ret;

	(void)usb_state_set_current_mode(USB_FUNCTION_NONE);
	change_usb_state_notification_handler(USB_FUNCTION_NONE);

	ret = usb_config_disable();
	if (ret < 0) {
		_E("Failed to disable USB config: %d", ret);
		return ret;
	}

	return 0;
}

/* Precondition: USB_DISCONNECTED (Implicitly contains USB_FUNCTION_NONE) */
static int usb_connected(void)
{
	int ret;
	unsigned int mode = usb_state_get_selected_mode();

	if (disp_plgn->pm_lock_internal)
		disp_plgn->pm_lock_internal(INTERNAL_LOCK_USB, LCD_OFF, STAY_CUR_STATE, 0);

	usb_state_set_connection(USB_CONNECTED);
	send_usb_state_changed_event(VCONFKEY_SYSMAN_USB_CONNECTED);

	ret = usb_enable(mode);
	if (ret < 0) {
		_E("Failed to enable USB gadget(%#x): %d", mode, ret);
		return ret;
	}

	return 0;
}

/* Precondition: USB_CONNECTED */
static int usb_disconnected(void)
{
	int ret;

	usb_state_set_connection(USB_DISCONNECTED);
	send_usb_state_changed_event(VCONFKEY_SYSMAN_USB_DISCONNECTED);

	ret = usb_disable();
	if(ret < 0) {
		_E("Failed to disable USB gadget: %d", ret);
		/* Important: You have to keep going to unlock disp_plgn->pm_unlock_internal */
	}

	if (disp_plgn->pm_unlock_internal)
		disp_plgn->pm_unlock_internal(INTERNAL_LOCK_USB, LCD_OFF, STAY_CUR_STATE);

	return ret;
}

/* Called by dbus signal and vconf change(tethering, debug mode) */
int usb_change_mode(unsigned int new_mode, bool change_debug_mode)
{
	int ret;
	const unsigned int curr_mode = usb_state_get_current_mode();
	const unsigned int prev_mode = usb_state_get_selected_mode();

	if (prev_mode == new_mode) {
		_I("already using USB mode(%#x), current running mode (%#x)", prev_mode, curr_mode);
		return 0;
	}

	_I("USB change mode (%#x) -> (%#x), current running mode (%#x)", prev_mode, new_mode, curr_mode);

	/*
	 * When you change the gadget(usb_change_gadget) for new usb mode, you must disable the gadget first.
	 *
	 * Even if the usb cable is plugged in, the current mode may be NULL due to usb fail.
	 * So to find out which mode you are in, you should use the current mode, not the selected mode.
	*/
	if (curr_mode != USB_FUNCTION_NONE) {
		/* Special case: Because usb gadget is disabled, temporarily switch usb state from available to connected. */
		send_usb_state_changed_event(VCONFKEY_SYSMAN_USB_CONNECTED);

		ret = usb_disable();
		if (ret < 0) {
			_E("Failed to disable current USB mode.");
			return ret;
		}
	}

	/*
	 * You should always change the gadget once when the usb mode is changed.
	 * In this state, usb_enable() is called when the usb cable is connected
	 * and usb_disable() is called when the usb cable is disconnected.
	 */
	ret = usb_change_gadget(new_mode);
	if (ret < 0) {
		_E("Failed to change USB gadget: %d", ret);
		return ret;
	}

	if (usb_state_get_connection() == USB_CONNECTED) {
		ret = usb_enable(new_mode);
		if (ret < 0) {
			_E("Failed to enable USB mode: %d", ret);
			return ret;
		}
	}

	/* If you success to change the runtime usb configuration, do change the selected mode. */
	(void)usb_state_set_selected_mode(new_mode);

	/* To sync with vconf for debug mode */
	if (change_debug_mode)
		set_usb_debug_state((bool)(new_mode & USB_FUNCTION_SDB));

	return 0;
}

/* Called by extcon udev event */
static int usb_state_changed(const char *index, int new_status)
{
	int ret = -1;
	static int old_status = -1;

	/* For debugging. Do not move the location. */
	_I("USB state is changed by extcon from (%d) to (%d).", old_status, new_status);

	if (old_status == new_status)
		return 0;

	switch (new_status) {
	case USB_CONNECTED:
		ret = usb_connected();
		break;

	case USB_DISCONNECTED:
		ret = usb_disconnected();
		break;

	default:
		_E("Invalid USB state(%d).", new_status);
		return -EINVAL;
	}

	if (ret < 0) {
		_E("Failed to operate USB connection: %d", ret);
		return ret;
	}

	old_status = new_status;

	return ret;
}

static void uevent_usb_mode_handler(struct udev_device *dev)
{
	const char *state = NULL;

	state = udev_device_get_property_value(dev, "USB_STATE");
	if (!state)
		return;

	if (strncmp(state, "CONFIGURED", strlen(state) + 1))
		return;

	_I("USB udev event happend : CONFIGURED USB_STATE");

	if (usb_state_get_selected_mode() & USB_FUNCTION_ACM)
		systemd_start_unit_wait_started ("data-router.service", NULL, -1);

	if (usb_state_get_selected_mode() & USB_FUNCTION_SDB)
		systemd_start_unit_wait_started ("sdbd.service", NULL, -1);

	if (usb_state_get_selected_mode() & USB_FUNCTION_MTP)
		systemd_start_unit_wait_started ("mtp-responder.service", NULL, -1);
}

static struct uevent_handler uh = {
	.subsystem = "usb_mode",
	.uevent_func = uevent_usb_mode_handler,
};

static void usb_init(void *data)
{
	int ret;

	usb_state_retrieve_selected_mode();

	ret = usb_probe();
	if (ret < 0) {
		_E("USB client cannot be used: %d", ret);
		return;
	}

	ret = usb_config_init();
	if (ret < 0)
		_E("Failed to initialize USB configuation.");

	ret = register_udev_uevent_control(&uh);
	if (ret < 0)
		_E("Failed to register udev event(%d) for USB", ret);

	ret = usb_dbus_init();
	if (ret < 0)
		_E("Failed to init dbus(%d) for USB", ret);

	add_usb_tethering_handler();
	add_usb_debug_handler();
}

static void usb_exit(void *data)
{
	remove_usb_debug_handler();
	remove_usb_tethering_handler();

	usb_state_set_connection(USB_DISCONNECTED);
	send_usb_state_changed_event(VCONFKEY_SYSMAN_USB_DISCONNECTED);
	(void)usb_state_set_current_mode(USB_FUNCTION_NONE);
	change_usb_state_notification_handler(USB_FUNCTION_NONE);

	unregister_udev_uevent_control(&uh);
	usb_config_deinit();
	usb_release();
}

static struct extcon_ops extcon_usb_ops = {
	.name	= EXTCON_CABLE_USB,
	.init	= usb_init,
	.exit	= usb_exit,
	.update = usb_state_changed,
};

EXTCON_OPS_REGISTER(extcon_usb_ops)

static void __CONSTRUCTOR__ initialize(void)
{
	disp_plgn = get_var_display_plugin();
	if (!disp_plgn)
		_E("Failed to get display plugin variable.");
}
